import numba
import numpy as np
import pandas as pd
from alphabase.constants._const import PEAK_INTENSITY_DTYPE, PEAK_MZ_DTYPE
from alphabase.io.hdf import HDF_File


@numba.njit
def _get_peaks_to_keep_mask(
    peak_count, peak_start_indices, peak_stop_indices, spec_mask
) -> tuple:
    """
    Generate a mask for peaks to keep and reindex spectrum indices

    Parameters
    ----------
    peak_count : int
        Total number of peaks in the dataset
    peak_start_indices : np.ndarray
        Array of peak start indices for each spectrum
    peak_stop_indices : np.ndarray
        Array of peak stop indices for each spectrum
    spec_mask : np.ndarray
        Boolean mask indicating which spectra to keep

    Returns
    -------
    tuple
        (peaks_mask, new_start_indices, new_stop_indices)
        - peaks_mask: boolean mask with True for peaks to keep
        - new_start_indices: reindexed start indices for each kept spectrum
        - new_stop_indices: reindexed stop indices for each kept spectrum
    """
    peaks_mask = np.zeros(peak_count, dtype=np.bool_)

    kept_spec_indices = np.nonzero(spec_mask)[0]

    for spec_idx in kept_spec_indices:
        start, stop = peak_start_indices[spec_idx], peak_stop_indices[spec_idx]
        peaks_mask[start:stop] = True

    new_start_indices = np.empty(len(kept_spec_indices), dtype=np.int64)
    new_stop_indices = np.empty(len(kept_spec_indices), dtype=np.int64)

    peak_cumsum = np.zeros(peak_count + 1, dtype=np.int64)
    for i in range(peak_count):
        peak_cumsum[i + 1] = peak_cumsum[i] + int(peaks_mask[i])

    for i, spec_idx in enumerate(kept_spec_indices):
        start, stop = peak_start_indices[spec_idx], peak_stop_indices[spec_idx]
        new_start_indices[i] = peak_cumsum[start]
        new_stop_indices[i] = peak_cumsum[stop]

    return peaks_mask, new_start_indices, new_stop_indices


class MSData_Base:
    """
    The base data structure for MS RAW Data, other MSData loaders inherit this class.
    """

    column_dtypes = {
        "rt": np.float64,
        "ms_level": np.int8,
        "precursor_mz": np.float64,
        "isolation_lower_mz": np.float64,
        "isolation_upper_mz": np.float64,
        "precursor_charge": np.int8,
        "nce": np.float32,
        "injection_time": np.float32,
        "activation": "U",
    }

    spectrum_df: pd.DataFrame
    """
    Spectrum dataframe containing the following columns:

    - `rt` (float64): in minutes. `rt_sec` will be RT in seconds, which is not included by default.
    - `precursor_mz` (float64): mono_mz (DDA) or isolation center mz
    - `isolation_lower_mz` (float64): left of the isolation window
    - `isolation_upper_mz` (float64): right of the isolation window
    - `spec_idx` (int64): spectrum index. For thermo, it is `scan_num - 1`
    - `peak_start_idx` (int64): peak start position pointing to `self.peak_df`
    - `peak_stop_idx` (int64): peak stop position pointing to `self._peak_df`
    - `ms_level` (int8): =1 for MS1, =2 for MS2, ...
    - [`scan_num`] (int64): Thermo scan number
    - [`mobility`] (float64): Bruker timsTOF mobility

    Other columns depends on different vendors and file formats
    """

    peak_df: pd.DataFrame
    """
    Peak list dataframe containing the follow columns:
    - `mz` (PEAK_MZ_DTYPE in alphabase, float32 by default): m/z values of the peak
    - `intensity` (PEAK_INTENSITY_DTYPE in alphabase, float32 by default): intensity values of the peak
    """

    vocab: list = [
        "CID",
        "HCD",
        "ETD",
        "ECD",
        "EAD",
        "EXD",
        "UVPD",
        "ETHCD",
        "ETCID",
        "EXCID",
        "NETD",
        "IT",
        "FT",
        "TOF",
    ]
    """
    These spectrum infomation items in str format can be one-to-one mapped into
    unique token IDs (indices), for exampel "CID"=0, "HCD"=1, ...
    Token IDs are better for storage in HDF5 format.
    """

    def __init__(self, centroided: bool = True, save_as_hdf: bool = False, **kwargs):
        """
        Parameters
        ----------
        centroided : bool, optional
            If peaks will be centroided after loading, by default True
        save_as_hdf : bool, optional
            If automatically save the data into HDF5 format, by default False
        """
        # A spectrum contains peaks
        self.spectrum_df: pd.DataFrame = pd.DataFrame()
        # A peak contains mz, intensity, and ...
        self.peak_df: pd.DataFrame = pd.DataFrame()
        self._raw_file_path = ""
        self.centroided = centroided
        self._save_as_hdf = save_as_hdf
        self.creation_time = ""
        self.file_type = ""
        self.instrument = "none"

    def _get_term_id(self, terminology: str) -> int:
        """
        Get terminology ID from :attr:`.MSData_Base.vocab`, -1 if not exist.

        Parameters
        ----------
        terminology : str
            The terminology name from :attr:`.MSData_Base.vocab`, such as "CID", "HCD", ...

        Returns
        -------
        int
            Terminology ID, which is the index in :attr:`.MSData_Base.vocab`.
        """
        try:
            return self.vocab.index(terminology)
        except ValueError:
            return -1

    @property
    def raw_file_path(self) -> str:
        return self._raw_file_path

    @raw_file_path.setter
    def raw_file_path(self, raw_file_path: str):
        self._raw_file_path = raw_file_path

    def import_raw(self, raw_file_path: str):
        """
        Import a raw file. It involves three steps:
        ```
        raw_data_dict = self._import(raw_file_path)
        self._set_dataframes(raw_data_dict)
        self._check_df()
        ```

        Parameters
        ----------
        raw_file_path : str
            Absolute or relative path of the raw file.
        """
        self.raw_file_path = raw_file_path
        raw_data_dict = self._import(raw_file_path)
        self._set_dataframes(raw_data_dict)
        self._check_df()

        if self._save_as_hdf:
            self.save_hdf(raw_file_path + ".hdf")

    def load_raw(self, raw_file_path: str):
        """
        Wrapper of :func:`.MSData_Base.import_raw`
        """
        self.import_raw(raw_file_path)

    def _save_meta_to_hdf(self, hdf: HDF_File):
        hdf.ms_data.meta = {
            "creation_time": self.creation_time,
            "raw_file_path": self.raw_file_path,
            "file_type": self.file_type,
            "centroided": self.centroided,
            "instrument": self.instrument,
        }

    def _load_meta_from_hdf(self, hdf: HDF_File):
        self.creation_time = hdf.ms_data.meta.creation_time
        self.raw_file_path = hdf.ms_data.meta.raw_file_path
        self.file_type = hdf.ms_data.meta.file_type
        self.centroided = hdf.ms_data.meta.centroided
        self.instrument = hdf.ms_data.meta.instrument

    def save_hdf(self, hdf_file_path: str):
        """
        Save data into HDF5 file

        Parameters
        ----------
        hdf_file_path : str
            Absolute or relative path of HDF5 file.
        """
        hdf = HDF_File(
            hdf_file_path, read_only=False, truncate=True, delete_existing=True
        )

        hdf.ms_data = {"spectrum_df": self.spectrum_df, "peak_df": self.peak_df}

        self._save_meta_to_hdf(hdf)

    def load_hdf(self, hdf_file_path: str):
        """
        Load data from HDF5 file.

        Parameters
        ----------
        hdf_file_path : str
            Absolute or relative path of HDF5 file.
        """
        hdf = HDF_File(
            hdf_file_path, read_only=True, truncate=False, delete_existing=False
        )

        self.spectrum_df = hdf.ms_data.spectrum_df.values
        self.peak_df = hdf.ms_data.peak_df.values

        if hasattr(hdf.ms_data, "meta"):
            self._load_meta_from_hdf(hdf)

    def reset_spec_idxes(self):
        """
        Reset spec indexes to make sure spec_idx values are continuous ranging from 0 to N.
        """
        self.spectrum_df.reset_index(drop=True, inplace=True)
        self.spectrum_df["spec_idx"] = self.spectrum_df.index.values

    def _import(self, _path: str) -> dict:
        """
        Parameters
        ----------
        _path : str
            Path of raw file.

        Returns
        -------
        dict
            Example:
            ```
            spec_dict = {
                "_peak_indices": _peak_indices,
                "peak_mz": np.concatenate(mz_values).copy(),
                "peak_intensity": np.concatenate(intensity_values).copy(),
                "rt": np.array(rt_values).copy(),
                "precursor_mz": np.array(precursor_mz_values).copy(),
                "precursor_charge": np.array(precursor_charges, dtype=np.int8).copy(),
                "isolation_lower_mz": np.array(isolation_mz_lowers).copy(),
                "isolation_upper_mz": np.array(isolation_mz_uppers).copy(),
                "ms_level": np.array(ms_order_list, dtype=np.int8).copy(),
                "nce": np.array(ce_list, dtype=np.float32).copy(),
            }
            ```

        Raises
        ------
        NotImplementedError
            Sub-class of `MSData_Base` must implement this method.
        """
        raise NotImplementedError(f"{self.__class__} must implement `_import()`")

    def _set_dataframes(self, raw_data: dict):
        self.create_spectrum_df(len(raw_data["rt"]))
        self.set_peak_df_by_indexed_array(
            raw_data["peak_mz"],
            raw_data["peak_intensity"],
            raw_data["peak_indices"][:-1],
            raw_data["peak_indices"][1:],
        )

        for col, val in raw_data.items():
            if col in self.column_dtypes:
                if self.column_dtypes[col] == "O":
                    self.spectrum_df[col] = list(val)
                else:
                    self.spectrum_df[col] = np.array(val, dtype=self.column_dtypes[col])

    def _read_creation_time(self, raw_data):
        pass

    def _check_df(self):
        self._check_rt()
        # self._check_mobility()
        self._check_precursor_mz()
        self._check_peak_dtypes()

    def _check_peak_dtypes(self):
        if self.peak_df.mz.dtype != PEAK_MZ_DTYPE:
            self.peak_df.mz = self.peak_df.mz.astype(PEAK_MZ_DTYPE)
        if self.peak_df.intensity.dtype != PEAK_INTENSITY_DTYPE:
            self.peak_df.intensity = self.peak_df.intensity.astype(PEAK_INTENSITY_DTYPE)

    def _check_rt(self):
        assert "rt" in self.spectrum_df.columns
        # self.spectrum_df['rt_sec'] = self.spectrum_df.rt*60

    def _check_mobility(self):
        if "mobility" not in self.spectrum_df.columns:
            self.spectrum_df["mobility"] = 0.0

    def _check_precursor_mz(self):
        if "isolation_lower_mz" not in self.spectrum_df.columns:
            self.spectrum_df["isolation_lower_mz"] = -1.0
            self.spectrum_df["isolation_upper_mz"] = -1.0
        if "precursor_mz" not in self.spectrum_df.columns:
            self.spectrum_df["precursor_mz"] = -1.0

    def create_spectrum_df(
        self,
        spectrum_num: int,
    ):
        """
        Create a empty spectrum dataframe from the number of spectra.

        Parameters
        ----------
        spectrum_num : int
            The number of spectra.
        """
        self.spectrum_df = pd.DataFrame(index=np.arange(spectrum_num, dtype=np.int64))
        self.spectrum_df["spec_idx"] = self.spectrum_df.index.values

    def set_peak_df_by_indexed_array(
        self,
        mz_array: np.ndarray,
        intensity_array: np.ndarray,
        peak_start_indices: np.ndarray,
        peak_stop_indices: np.ndarray,
    ):
        self.peak_df = pd.DataFrame(
            {
                "mz": mz_array.astype(PEAK_MZ_DTYPE),
                "intensity": intensity_array.astype(PEAK_INTENSITY_DTYPE),
            }
        )
        self.spectrum_df["peak_start_idx"] = peak_start_indices
        self.spectrum_df["peak_stop_idx"] = peak_stop_indices

    def set_peak_df_by_array_list(
        self,
        mz_array_list: list,
        intensity_array_list: list,
    ):
        indices = index_ragged_list(mz_array_list)
        self.set_peak_df_by_indexed_array(
            np.concatenate(mz_array_list),
            np.concatenate(intensity_array_list),
            indices[:-1],
            indices[1:],
        )

    def add_column_in_spec_df_by_spec_idxes(
        self,
        column_name: str,
        values: np.ndarray,
        spec_idxes: np.ndarray,
        dtype: np.dtype = np.float64,
        na_value=np.nan,
    ):
        self.spectrum_df.loc[spec_idxes, column_name] = values
        self.spectrum_df[column_name].fillna(na_value, inplace=True)
        self.spectrum_df[column_name] = self.spectrum_df[column_name].astype(dtype)

    def add_column_in_df_by_scan_num(
        self,
        column_name: str,
        values: np.ndarray,
        scan_nums: np.ndarray,
        dtype: np.dtype = np.float64,
        na_value=np.nan,
    ):
        """
        scan num starts from 1 not 0
        """
        self.add_column_in_spec_df_by_spec_idxes(
            column_name, values, scan_nums - 1, dtype, na_value
        )

    def get_peaks(self, spec_idx):
        start, end = self.spectrum_df[["peak_start_idx", "peak_stop_idx"]].values[
            spec_idx, :
        ]
        return (
            self.peak_df.mz.values[start:end],
            self.peak_df.intensity.values[start:end],
        )

    def set_precursor_mz_by_spec_idxes(
        self,
        precursor_mz_values: np.ndarray,
        spec_idxes: np.ndarray,
    ):
        self.add_column_in_spec_df_by_spec_idxes(
            "precursor_mz", precursor_mz_values, spec_idxes, np.float64, -1.0
        )

    def set_isolation_mz_windows_by_spec_idxes(
        self,
        precursor_lower_mz_values: np.ndarray,
        precursor_upper_mz_values: np.ndarray,
        spec_idxes: np.ndarray,
    ):
        self.add_column_in_spec_df_by_spec_idxes(
            "isolation_lower_mz",
            precursor_lower_mz_values,
            spec_idxes,
            np.float64,
            -1.0,
        )
        self.add_column_in_spec_df_by_spec_idxes(
            "isolation_upper_mz",
            precursor_upper_mz_values,
            spec_idxes,
            np.float64,
            -1.0,
        )

    def _sort_rt(self):
        """
        Used by :class:`alpharaw.wrappers.alphatims_wrapper.AlphaTims_Wrapper`.
        """
        # if 'mobility' in self.spectrum_df.columns:
        #     self.spectrum_df['_mobility'] = -self.spectrum_df.mobility
        #     self.spectrum_df.sort_values(['rt','_mobility'],inplace=True)
        #     self.spectrum_df.drop(columns=['_mobility'],inplace=True)
        # else:
        #     self.spectrum_df.sort_values('rt',inplace=True)
        self.spectrum_df.sort_values("rt", inplace=True)
        self.spectrum_df.reset_index(drop=True, inplace=True)
        self.spectrum_df["spec_idx_old"] = self.spectrum_df.spec_idx
        self.spectrum_df["spec_idx"] = self.spectrum_df.index
        mzs_list = []
        intens_list = []
        idx_list = []
        for start, end in self.spectrum_df[["peak_start_idx", "peak_stop_idx"]].values:
            mzs_list.append(self.peak_df.mz.values[start:end])
            intens_list.append(self.peak_df.intensity.values[start:end])
            idx_list.append(end - start)
        peak_indices = np.empty(len(idx_list) + 1, dtype=np.int64)
        peak_indices[0] = 0
        peak_indices[1:] = np.cumsum(idx_list)
        self.peak_df.mz.values[:] = np.concatenate(mzs_list)
        self.peak_df.intensity.values[:] = np.concatenate(intens_list)
        self.spectrum_df["peak_start_idx"] = peak_indices[:-1]
        self.spectrum_df["peak_stop_idx"] = peak_indices[1:]

    def remove_unused_peaks(self):
        """

        If the spectrum_df object is filtered, the spectra are not removed from the peak_df.
        This method will remove the unused peaks from the peak_df and reindex the peak indices in spectrum_df.

        Assumes that spectra have already been removed from spectrum_df.
        This method will rebuild the peak_df to contain only referenced peaks.
        Preserves all columns in both peak_df and spectrum_df.

        Returns
        -------
        MSData_Base or None
            New instance if in_place=False, None otherwise.
        """
        spec_mask = np.ones(len(self.spectrum_df), dtype=bool)
        peaks_mask, new_start_indices, new_stop_indices = _get_peaks_to_keep_mask(
            len(self.peak_df),
            self.spectrum_df.peak_start_idx.values,
            self.spectrum_df.peak_stop_idx.values,
            spec_mask,
        )

        new_peak_df = self.peak_df[peaks_mask].reset_index(drop=True)

        self.peak_df = new_peak_df
        self.spectrum_df["peak_start_idx"] = new_start_indices
        self.spectrum_df["peak_stop_idx"] = new_stop_indices
        return None


def index_ragged_list(ragged_list: list) -> np.ndarray:
    """Create lookup indices for a list of arrays for concatenation.

    Args:
        value (list): Input list of arrays.

    Returns:
        indices: A numpy array with indices.
    """
    indices = np.zeros(len(ragged_list) + 1, np.int64)
    indices[1:] = [len(i) for i in ragged_list]
    indices = np.cumsum(indices, dtype=np.int64)

    return indices


class MSData_HDF(MSData_Base):
    """
    Wrapper of reader for alpharaw's HDF5 spectrum file.
    This class will be registered as file formats (types)
    "alpharaw", "raw.hdf", "alpharaw_hdf", "hdf" and "hdf5"
    in :obj:`ms_reader_provider` instance by :func:`register_readers`.
    """

    def import_raw(self, _path: str):
        self.raw_file_path = _path
        self.load_hdf(_path)


class MSReaderProvider:
    """Factory class to register and get MS Readers"""

    def __init__(self):
        self.ms_reader_dict = {}

    def register_reader(self, ms_file_type: str, reader_class: type):
        """
        Register a new reader for `ms_file_type` format with `reader_class`.

        Parameters
        ----------
        ms_file_type : str
            AlphaRaw supported MS file types.
        reader_class : type
            AlphaRaw supported MS class types.
        """
        self.ms_reader_dict[ms_file_type.lower()] = reader_class

    def get_reader(
        self, ms_file_type: str, *, centroided: bool = True, **kwargs
    ) -> MSData_Base:
        """
        Get the MS reader for the given `ms_file_type`.

        Parameters
        ----------
        ms_file_type : str
            AlphaRaw supported MS file types.
        centroided : bool, optional
            If peaks will be centroided after loading, by default True.

        Returns
        -------
        MSData_Base
            Instance of corresponding sub-class of `MSData_Base`.
        """
        ms_file_type = ms_file_type.lower()
        if ms_file_type not in self.ms_reader_dict:
            return None
        else:
            return self.ms_reader_dict[ms_file_type](centroided=centroided, **kwargs)


ms_reader_provider = MSReaderProvider()
"""
MS data register (:class:`MSReaderProvider`) performs as a factory to
produce different readers for given file formats.
"""


def register_readers():
    """
    Register :class:`MSData_HDF` for file formats (types):
    "alpharaw", "raw.hdf", "alpharaw_hdf", "hdf", "hdf5"
    in :obj:`ms_reader_provider`.
    """
    ms_reader_provider.register_reader("alpharaw", MSData_HDF)
    ms_reader_provider.register_reader("raw.hdf", MSData_HDF)
    ms_reader_provider.register_reader("alpharaw_hdf", MSData_HDF)
    ms_reader_provider.register_reader("hdf", MSData_HDF)
    ms_reader_provider.register_reader("hdf5", MSData_HDF)


register_readers()



import typing
from collections import defaultdict

import numpy as np
from alphatims.bruker import TimsTOF

from alpharaw.ms_data_base import MSData_Base
from alpharaw.utils.df_processing import remove_unused_peaks
from alpharaw.utils.timstof import convert_to_alphatims


class NormalDIAGrouper:
    def __init__(self, ms_data: MSData_Base):
        self.ms_data = ms_data
        self.ms_data.spectrum_df["dia_group"] = ms_data.spectrum_df.precursor_mz.astype(
            int
        )

        self.dia_group_dfs = self.ms_data.spectrum_df.groupby("dia_group")
        self.dia_isolation_dict = {}
        for dia_group, df in self.dia_group_dfs:
            if dia_group == -1:
                continue
            self.dia_isolation_dict[dia_group] = (
                df.isolation_lower_mz.values[0],
                df.isolation_upper_mz.values[0],
            )
        self.dia_groups = np.sort(list(self.dia_isolation_dict.keys()))

    def get_ms_data_for_a_group(
        self,
        dia_group: int = -1,
        return_alpharaw_data: bool = True,
        return_alphatims_data: bool = True,
    ) -> typing.Union[MSData_Base, TimsTOF, typing.Tuple[MSData_Base, TimsTOF]]:
        """Get compressed MS data for isolation window `dia_group`.

        Args:
            dia_group (int, optional): The DIA group, -1 means ms1. Defaults to -1.
            return_alphatims_data (bool, optional): If return `MSData_Base`. Defaults to True
            return_alphatims_data (bool, optional): If return alphatims object. Defaults to True.

        Returns:
            MSData_Base: Compressed MS data, if `return_alpharaw_data==True`
            TimsTOF: Alphatims object for the window, if `return_alphatims_data==True`
        """

        spec_df = self.dia_group_dfs.get_group(dia_group)

        if return_alphatims_data:
            ms_data, ms_tims = convert_to_alphatims(
                spec_df, self.ms_data.peak_df, dda=False
            )
            if return_alpharaw_data:
                return ms_data, ms_tims
            else:
                return ms_tims
        else:
            ms_data = MSData_Base()

            spec_df, peak_df = remove_unused_peaks(spec_df, self.ms_data.peak_df)

            ms_data.spectrum_df = spec_df
            ms_data.peak_df = peak_df
            return ms_data

    def assign_dia_groups(
        self, precursor_mzs
    ) -> typing.DefaultDict[typing.List, typing.List]:
        dia_precursor_groups = defaultdict(list)
        for i, mz in enumerate(precursor_mzs):
            i_group = np.searchsorted(self.dia_groups, int(mz))
            if i_group == 0:
                i_group = 1
            elif i_group == len(self.dia_groups):
                i_group -= 1
            dia_group = self.dia_groups[i_group]
            if dia_group in self.dia_isolation_dict:
                isolation_lower, isolation_upper = self.dia_isolation_dict[dia_group]
                if mz >= isolation_lower and mz <= isolation_upper:
                    dia_precursor_groups[dia_group].append(i)
                    continue
            dia_group = self.dia_groups[i_group - 1]
            if dia_group in self.dia_isolation_dict:
                isolation_lower, isolation_upper = self.dia_isolation_dict[dia_group]
                if mz >= isolation_lower and mz <= isolation_upper:
                    dia_precursor_groups[dia_group].append(i)
        return dia_precursor_groups



import numpy as np
from alphabase.io.hdf import HDF_File

from alpharaw.ms_data_base import MSData_Base, ms_reader_provider


class AlphaPept_HDF_MS2_Reader(MSData_Base):
    """MS2 from AlphaPept HDF"""

    def _import(self, _path):
        return _path

    def _set_dataframes(self, _path):
        hdf = HDF_File(_path)
        self.peak_df["mz"] = hdf.Raw.MS2_scans.mass_list_ms2.values
        self.peak_df["intensity"] = hdf.Raw.MS2_scans.int_list_ms2.values
        if hasattr(hdf.Raw.MS2_scans, "mobility2"):
            spec_idxes = np.arange(len(hdf.Raw.MS2_scans.rt_list_ms2), dtype=np.int64)
        else:
            spec_idxes = hdf.Raw.MS2_scans.scan_list_ms2.values - 1

        spec_num = spec_idxes.max() + 1
        self.create_spectrum_df(spec_num)

        peak_indices = hdf.Raw.MS2_scans.indices_ms2.values
        start_idxes = np.full(spec_num, -1, dtype=np.int64)
        start_idxes[spec_idxes] = peak_indices[:-1]
        end_idxes = np.full(spec_num, -1, dtype=np.int64)
        end_idxes[spec_idxes] = peak_indices[1:]

        self.set_peak_df_by_indexed_array(
            hdf.Raw.MS2_scans.mass_list_ms2.values,
            hdf.Raw.MS2_scans.int_list_ms2.values,
            start_idxes,
            end_idxes,
        )

        self.add_column_in_spec_df_by_spec_idxes(
            "rt",
            hdf.Raw.MS2_scans.rt_list_ms2.values,
            spec_idxes,
            dtype=np.float64,
            na_value=0,
        )
        self.add_column_in_spec_df_by_spec_idxes(
            "ms_level", 2, spec_idxes, dtype=np.int8, na_value=1
        )

        self.add_column_in_spec_df_by_spec_idxes(
            "charge",
            hdf.Raw.MS2_scans.charge2.values,
            spec_idxes,
            dtype=np.int8,
            na_value=0,
        )

        if hasattr(hdf.Raw.MS2_scans, "mobility2"):
            self.add_column_in_spec_df_by_spec_idxes(
                "mobility",
                hdf.Raw.MS2_scans.mobility2.values,
                spec_idxes,
            )

        if hasattr(hdf.Raw.MS2_scans, "mono_mzs2"):
            self.add_column_in_spec_df_by_spec_idxes(
                "precursor_mz",
                hdf.Raw.MS2_scans.mono_mzs2.values,
                spec_idxes,
                na_value=-1.0,
            )
            # self.set_isolation_mz_windows(
            #     precursor_mzs-1.5, precursor_mzs+1.5
            # )


ms_reader_provider.register_reader("alphapept", AlphaPept_HDF_MS2_Reader)
ms_reader_provider.register_reader("alphapept_hdf", AlphaPept_HDF_MS2_Reader)


def get_peak_lists(starts, ends, peak_df):
    mass_list = [
        peak_df.mz.values[start:end].tolist() for start, end in zip(starts, ends)
    ]
    inten_list = [
        peak_df.intensity.values[start:end].tolist() for start, end in zip(starts, ends)
    ]
    return mass_list, inten_list


def extract_ms1(raw_data: MSData_Base, query_data: dict):
    spec_df = raw_data.spectrum_df.query("ms_level==1")
    scans = spec_df.spec_idx.values
    rts = spec_df.rt.values
    ms_levels = spec_df.ms_level.values

    mass_list_ms1, int_list_ms1 = get_peak_lists(
        spec_df.peak_start_idx.values, spec_df.peak_stop_idx.values, raw_data.peak_df
    )

    query_data["scan_list_ms1"] = scans
    query_data["rt_list_ms1"] = rts
    query_data["mass_list_ms1"] = np.array(mass_list_ms1, dtype=object)
    query_data["int_list_ms1"] = np.array(int_list_ms1, dtype=object)
    query_data["ms_list_ms1"] = ms_levels


def extract_ms2(raw_data: MSData_Base, query_data: dict):
    spec_df = raw_data.spectrum_df.query("ms_level==2")
    scans = spec_df.spec_idx.values
    rts = spec_df.rt.values
    ms_levels = spec_df.ms_level.values
    precursor_mzs2 = spec_df.precursor_mz.values
    charges = spec_df.charge.values
    charges[charges <= 0] = 2

    mass_list_ms2, int_list_ms2 = get_peak_lists(
        spec_df.peak_start_idx.values, spec_df.peak_stop_idx.values, raw_data.peak_df
    )

    query_data["scan_list_ms2"] = scans
    query_data["rt_list_ms2"] = rts
    query_data["mass_list_ms2"] = mass_list_ms2
    query_data["int_list_ms2"] = int_list_ms2
    query_data["ms_list_ms2"] = ms_levels
    query_data["prec_mass_list2"] = precursor_mzs2
    query_data["mono_mzs2"] = precursor_mzs2
    query_data["charge2"] = charges


def parse_msdata_to_alphapept(raw_data: MSData_Base):
    query_data = {}
    extract_ms1(raw_data, query_data)
    extract_ms2(raw_data, query_data)

    return query_data, raw_data.creation_time



import typing

import pandas as pd
from alphatims.bruker import TimsTOF

from alpharaw.ms_data_base import MSData_Base
from alpharaw.utils.df_processing import remove_unused_peaks
from alpharaw.wrappers.alphatims_wrapper import AlphaTimsWrapper


def convert_to_alphatims(
    spec_df: pd.DataFrame,
    peak_df: pd.DataFrame,
    dda: bool = False,
) -> typing.Tuple[MSData_Base, TimsTOF]:
    """
    Convert any spectrum dataframe or sliced spectrum dataframe
    and its peak dataframe into AlphaTims' TimsTOF object (AlphaTimsWrapper).

    Args:
        spec_df (pd.DataFrame):
            spectrum dataframe or sliced spectrum dataframe in AlphaRaw's format.
        peak_df (pd.DataFrame):
            peak dataframe in AlphaRaw's format by removing unused peaks in spec_df.
        dda (bool):
            if dda data.

    Returns:
        MSData_Base: AlphaRaw object
        TimsTOF: AlphaTims' TimsTOF object (AlphaTimsWrapper).
    """
    spec_df, peak_df = remove_unused_peaks(spec_df, peak_df)
    ms_data = MSData_Base()
    ms_data.spectrum_df = spec_df
    ms_data.peak_df = peak_df

    return ms_data, AlphaTimsWrapper(ms_data, dda=dda)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import importlib
import inspect
import os
import sys

sys.path.insert(0, os.path.abspath(".."))


# -- Project information -----------------------------------------------------

project = "alpharaw"
copyright = "2022, Mann Labs, MPIB"
author = "Mann Labs, MPIB"

release = "0.4.7"

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx.ext.linkcode",
    "sphinx.ext.viewcode",
    # 'sphinx.ext.autodoc',
    "autodocsumm",
    "nbsphinx",
    "myst_parser",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "_modidx,py"]

code_url = "https://github.com/mannlabs/alpharaw/blob/main"


def linkcode_resolve(domain, info):
    # Non-linkable objects from the starter kit in the tutorial.
    if domain == "js" or info["module"] == "connect4":
        return

    if domain != "py":
        return

    mod = importlib.import_module(info["module"])
    if "." in info["fullname"]:
        objname, attrname = info["fullname"].split(".")
        obj = getattr(mod, objname)
        try:
            # object is a method of a class
            obj = getattr(obj, attrname)
        except AttributeError:
            # object is an attribute of a class
            return None
    else:
        obj = getattr(mod, info["fullname"])

    try:
        file = inspect.getsourcefile(obj)
        lines = inspect.getsourcelines(obj)
    except TypeError:
        # e.g. object is a typing.Union
        return None
    file = os.path.relpath(file, os.path.abspath(".."))
    if not file.startswith("alpharaw"):
        # e.g. object is a typing.NewType
        return None
    start, end = lines[1], lines[1] + len(lines[0]) - 1

    return f"{code_url}/{file}#L{start}-L{end}"


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_css_files = [
    "css/custom.css",
]

autodoc_default_options = {
    "autosummary": True,
    "special-members": "__init__",  # Include __init__ methods.
}



import os

import click

import alpharaw
from alpharaw.legacy_msdata import mgf  # noqa: F401  # TODO remove import side effect
from alpharaw.ms_data_base import ms_reader_provider
from alpharaw.mzml import MzMLReader  # noqa: F401  # TODO remove import side effect
from alpharaw.wrappers import (
    alphapept_wrapper,  # noqa: F401  # TODO remove import side effect
)

try:
    from alpharaw.sciex import (
        SciexWiffData,  # noqa: F401 # TODO remove import side effect
    )
    from alpharaw.thermo import (
        ThermoRawData,  # noqa: F401 # TODO remove import side effect
    )
except (RuntimeError, ImportError):
    print("[WARN] pythonnet is not installed")


@click.group(
    context_settings=dict(
        help_option_names=["-h", "--help"],
    ),
    invoke_without_command=True,
)
@click.pass_context
@click.version_option(alpharaw.__version__, "-v", "--version")
def run(ctx, **kwargs):
    click.echo(
        rf"""
   ___   __     __        ___
  / _ | / /__  / /  ___ _/ _ \___ __    __
 / __ |/ / _ \/ _ \/ _ `/ , _/ _ `/ |/|/ /
/_/ |_/_/ .__/_//_/\_,_/_/|_|\_,_/|__,__/
       /_/
....................................................
.{alpharaw.__version__.center(50)}.
.{alpharaw.__github__.center(50)}.
.{alpharaw.__license__.center(50)}.
....................................................
"""
    )
    if ctx.invoked_subcommand is None:
        click.echo(run.get_help(ctx))


@run.command("parse", help="Convert raw files into alpharaw hdf5 (.hdf) format.")
@click.option(
    "--raw_type",
    type=str,
    default="thermo_raw",
    show_default=True,
    help="Only `thermo_raw`, `sciex_wiff` is supported currently.",
)
@click.option(
    "--raw",
    multiple=True,
    default=[],
    show_default=True,
    help="Raw files, can be `--raw raw1 --raw raw2 ...`.",
)
def _parse(raw_type: str, raw: list):
    reader = ms_reader_provider.get_reader(raw_type)
    if reader is None:
        print(
            f"{raw_type} is not supported, this may be due to the failed installion of PythonNet or other packages"
        )
    else:
        for raw_file in raw:
            if not os.path.isfile(raw_file):
                print(f"{raw_file} does not exist")
                continue
            reader.import_raw(raw_file)



import multiprocessing as mp
import platform
from functools import partial

import numpy as np
from tqdm import tqdm

import alpharaw.raw_access.pythermorawfilereader as pyrawfilereader

from .ms_data_base import (
    PEAK_INTENSITY_DTYPE,
    PEAK_MZ_DTYPE,
    MSData_Base,
    ms_reader_provider,
)

#: These thermo spectrum items can be only accessed by trailer dict using RawFileReader APIs.
__trailer_extra_list__ = [
    "injection_time",
    "cv",
    "max_ion_time",
    "agc_target",
    "energy_ev",
    "injection_optics_settling_time",
    "funnel_rf_level",
    "faims_cv",
    "scan_node",
]

auxiliary_item_dtypes: dict = {
    "injection_time": np.float32,
    "cv": np.float32,
    "max_ion_time": np.float32,
    "agc_target": np.int64,
    "energy_ev": np.float32,
    "injection_optics_settling_time": np.float32,
    "funnel_rf_level": np.float32,
    "faims_cv": np.float32,
    "scan_node": "U",
    "detector": "U",
    "activation": "U",
    "analyzer": "U",
    "detector_id": np.uint8,
    "activation_id": np.uint8,
    "analyzer_id": np.uint8,
    "scan_event_string": "U",
    "multinotch": "O",
}
"""
The auxiliary items and dtypes that can be accessed from thermo RawFileReader.
"""


class ThermoRawData(MSData_Base):
    """
    Loading Thermo Raw data as :class:`alpharaw.ms_data_base.MSData_Base` data structure.
    This class will be registered as file formats "thermo" and "thermo_raw" in
    :obj:`alpharaw.ms_data_base.ms_reader_provider` by local :func:`register_readers`.
    """

    MASS_ANALYZER_ID_TO_TYPE = dict(
        (_id, _t)
        for _id, _t in pyrawfilereader.RawFileReader.massAnalyzerType.items()
        if isinstance(_id, int)
    )
    """
    The dict of Thermo's mass analyzer ID (int) to its name (str).
    The IDs correspond to enum values (int) of `int(scan_event.MassAnalyzer)`
    in RawFileReader, which originate from
    `ThermoFisher.CommonCore.Data.FilterEnums.MassAnalyzerType`.
    """

    MASS_ANALYZER_TYPE_TO_ID = dict(
        (_t, _id)
        for _t, _id in pyrawfilereader.RawFileReader.massAnalyzerType.items()
        if isinstance(_id, int)
    )
    """
    The dict of Thermo's mass analyzer name (str) to its ID (int).
    The IDs correspond to enum values of `int(scan_event.MassAnalyzer)`
    in RawFileReader, which originate from
    `ThermoFisher.CommonCore.Data.FilterEnums.MassAnalyzerType`.
    """

    ACTIVATION_ID_TO_TYPE = dict(
        (_id, _t)
        for _id, _t in pyrawfilereader.RawFileReader.activationType.items()
        if isinstance(_id, int)
    )
    """
    The dict of Thermo's activation name (str) to its ID (int).
    The IDs correspond to enum values (int) of `int(scan_event.GetActivation(index))`
    in RawFileReader, which originate from
    `ThermoFisher.CommonCore.Data.FilterEnums.ActivationType`.
    Note that multiple activations like `ETHCD` and `ETCID` are
    not thermo's built-in activation types,
    AlphaRaw still assigns IDs to them (ETHCD=201, ETCID=202).
    We can get multiple activations from auxiliary_item `scan_event_string`.
    """

    ACTIVATION_TYPE_TO_ID = dict(
        (_t, _id)
        for _t, _id in pyrawfilereader.RawFileReader.activationType.items()
        if isinstance(_id, int)
    )
    """
    The dict of Thermo's activation ID (int) to its name (str).
    The IDs correspond to enum values (int) of `int(scan_event.GetActivation(index))`
    in RawFileReader, which originate from
    `ThermoFisher.CommonCore.Data.FilterEnums.ActivationType`.
    Note that `ETHCD` and `ETCID` are not thermo's built-in activation types,
    AlphaRaw still assigns IDs to them (ETHCD=201, ETCID=202).
    We can get multiple activations from auxiliary_item `scan_event_string`.
    """

    def __init__(
        self,
        centroided: bool = True,
        process_count: int = 10,
        mp_batch_size: int = 5000,
        save_as_hdf: bool = False,
        dda: bool = False,
        auxiliary_items: list = [],
        **kwargs,
    ):
        """
        Parameters
        ----------
        centroided : bool, optional
            If peaks will be centroided after loading. By defaults True.
        process_count : int, optional
            Number of processes to load RAW data, by default 10.
        mp_batch_size : int, optional
            Number of spectra to load in each batch, by default 5000.
        save_as_hdf : bool, optional
            Automatically save hdf after load raw data, by default False.
        dda : bool, optional
            Is DDA data, by default False.
        auxiliary_items : list, optional
            Additional spectrum items, candidates are in :data:`auxiliary_item_dtypes`.
            By default [].
        """
        super().__init__(centroided, save_as_hdf=save_as_hdf, **kwargs)
        self.file_type = "thermo"
        self.process_count = process_count
        self.mp_batch_size = mp_batch_size
        self.dda = dda
        self.auxiliary_items = auxiliary_items
        self.column_dtypes.update(auxiliary_item_dtypes)

    def _import(
        self,
        raw_file_path: str,
    ) -> dict:
        """
        Re-implementation of :func:`MSData_Base._import` to enable :func:`.MSData_Base.import_raw`.

        Parameters
        ----------
        raw_file_path : str
            File path of the raw data.

        Returns
        -------
        dict
            Spectrum information in a temporary dict format.
        """
        rawfile = pyrawfilereader.RawFileReader(raw_file_path)
        self.creation_time = rawfile.GetCreationDate()

        # create batches for multiprocessing
        first_spectrum_number = rawfile.FirstSpectrumNumber
        last_spectrum_number = rawfile.LastSpectrumNumber

        if self.process_count > 1:
            mode = "spawn" if platform.system() != "Linux" else "forkserver"

            batches = np.arange(
                first_spectrum_number, last_spectrum_number + 1, self.mp_batch_size
            )
            batches = np.append(batches, last_spectrum_number + 1)

            # use multiprocessing to load batches
            _import_batch_partial = partial(
                _import_batch,
                raw_file_path=raw_file_path,
                centroided=self.centroided,
                dda=self.dda,
                auxiliary_items=self.auxiliary_items,
            )
            with mp.get_context(mode).Pool(processes=self.process_count) as pool:
                batches = list(
                    tqdm(
                        pool.imap(_import_batch_partial, zip(batches[:-1], batches[1:]))
                    )
                )

            # collect peak indices
            _peak_indices = np.concatenate(
                [batch["_peak_indices"] for batch in batches]
            )
            peak_indices = np.empty(rawfile.LastSpectrumNumber + 1, np.int64)
            peak_indices[0] = 0
            peak_indices[1:] = np.cumsum(_peak_indices)

            output_dict = {"peak_indices": peak_indices}

            # concatenate other arrays
            for key in batches[0]:
                if key == "_peak_indices":
                    continue
                output_dict[key] = np.concatenate([batch[key] for batch in batches])
        else:
            output_dict = _import_batch(
                (first_spectrum_number, last_spectrum_number + 1),
                raw_file_path,
                self.centroided,
                dda=self.dda,
                auxiliary_items=self.auxiliary_items,
            )
            peak_indices = np.empty(rawfile.LastSpectrumNumber + 1, np.int64)
            peak_indices[0] = 0
            peak_indices[1:] = np.cumsum(output_dict["_peak_indices"])
            output_dict["peak_indices"] = peak_indices
        rawfile.Close()

        return output_dict


def _import_batch(
    start_stop_tuple: tuple,
    raw_file_path: str,
    centroided: bool,
    dda: bool,
    auxiliary_items: list,
) -> dict:
    """Collect spectra from a batch of scans.

    Parameters
    ----------
    start_stop_tuple : tuple
        tuple of start and stop scan numbers

    raw_file_path : str
        path to the raw file

    centroided : bool
        if centroided peaks should be collected

    dda : bool
        is dda data.

    auxiliary_items : list
        Candidates are in :data:`auxiliary_item_dtypes`.

    Returns
    -------
    dict
        dictionary of collected spectra

    """

    rawfile = pyrawfilereader.RawFileReader(raw_file_path)
    start, stop = start_stop_tuple
    use_trailer_extra = False

    _peak_indices = []
    mz_values = []
    intensity_values = []
    rt_values = []
    precursor_mz_values = []
    isolation_mz_lowers = []
    isolation_mz_uppers = []
    precursor_charges = []
    ms_order_list = []
    ce_list = []

    auxiliary_dict = dict((item, []) for item in auxiliary_items)
    if dda:
        use_trailer_extra = True
    else:
        for item in __trailer_extra_list__:
            if item in auxiliary_dict:
                use_trailer_extra = True
                break

    for i in range(start, stop):
        if not centroided:
            masses, intensities = rawfile.GetProfileMassListFromScanNum(i)
        else:
            masses, intensities = rawfile.GetCentroidMassListFromScanNum(i)
        mz_values.append(masses.astype(PEAK_MZ_DTYPE))
        intensity_values.append(intensities.astype(PEAK_INTENSITY_DTYPE))
        _peak_indices.append(len(masses))
        rt = rawfile.RTFromScanNum(i)
        rt_values.append(rt)
        ms_order = rawfile.GetMSOrderForScanNum(i)
        ms_order_list.append(ms_order)

        if use_trailer_extra:
            trailer_data = rawfile.GetTrailerExtraForScanNum(i)
        scan_event = rawfile.GetScanEventForScanNum(i)

        if "injection_time" in auxiliary_dict:
            auxiliary_dict["injection_time"].append(
                float(trailer_data["Ion Injection Time (ms):"])
            )
        if "scan_node" in auxiliary_dict:
            # String indicating experiment type in Thermo tune setup (e.g., "DIA", "Full scan"), helps differentiate scans with same MS order but from different experiments.
            # "Scan Node" is not a typo but hows its listed in the Trailer
            auxiliary_dict["scan_node"].append(
                _safe_get_trailer_value(trailer_data, "Scan Node:", "")
            )
        if "max_ion_time" in auxiliary_dict:
            auxiliary_dict["max_ion_time"].append(
                float(trailer_data["Max. Ion Time (ms):"])
            )
        if "agc_target" in auxiliary_dict:
            auxiliary_dict["agc_target"].append(float(trailer_data["AGC Target:"]))
        if "energy_ev" in auxiliary_dict:
            energy_ev = trailer_data["HCD Energy V:"]
            if not energy_ev:
                energy_ev = trailer_data["HCD Energy eV:"]
            if not energy_ev:
                energy_ev = 0
            auxiliary_dict["energy_ev"].append(float(energy_ev))
        if "injection_optics_settling_time" in auxiliary_dict:
            auxiliary_dict["injection_optics_settling_time"].append(
                float(trailer_data["Injection Optics Settling Time (ms):"])
            )
        if "funnel_rf_level" in auxiliary_dict:
            auxiliary_dict["funnel_rf_level"].append(
                float(trailer_data["Funnel RF Level:"])
            )
        if "faims_cv" in auxiliary_dict:
            auxiliary_dict["faims_cv"].append(float(trailer_data["FAIMS CV:"]))
        if "activation" in auxiliary_dict:
            auxiliary_dict["activation"].append(
                ThermoRawData.ACTIVATION_ID_TO_TYPE[int(scan_event.GetActivation(0))]
                if ms_order > 1
                else "MS1"
            )
        if "activation_id" in auxiliary_dict:
            auxiliary_dict["activation_id"].append(
                int(scan_event.GetActivation(0)) if ms_order > 1 else 255
            )
        if "analyzer" in auxiliary_dict:
            auxiliary_dict["analyzer"].append(
                ThermoRawData.MASS_ANALYZER_ID_TO_TYPE[int(scan_event.MassAnalyzer)]
            )
        if "analyzer_id" in auxiliary_dict:
            auxiliary_dict["analyzer_id"].append(int(scan_event.MassAnalyzer))
        if "scan_event_string" in auxiliary_dict:
            auxiliary_dict["scan_event_string"].append(
                rawfile.GetScanEventStringForScanNum(i)
            )
        if "multinotch" in auxiliary_dict:
            windows = []
            for _ in range(scan_event.MassCount):
                _mass = scan_event.GetMass(_)
                _width = scan_event.GetIsolationWidth(_)
                windows.append((_mass - _width / 2, _mass + _width / 2))
            auxiliary_dict["multinotch"].append(np.array(windows))

        if ms_order == 1:
            ce_list.append(0)
            if "cv" in auxiliary_dict:
                auxiliary_dict["cv"].append(0)
            precursor_mz_values.append(-1.0)
            isolation_mz_lowers.append(-1.0)
            isolation_mz_uppers.append(-1.0)
            precursor_charges.append(0)
        else:
            ce_list.append(rawfile.GetCollisionEnergyForScanNum(i))
            if "cv" in auxiliary_dict:
                n_cvs = scan_event.SourceFragmentationInfoCount
                if n_cvs > 0:
                    auxiliary_dict["cv"].append(
                        scan_event.GetSourceFragmentationInfo(0)
                    )
                else:
                    auxiliary_dict["cv"].append(0)

            isolation_center = scan_event.GetReaction(0).PrecursorMass
            isolation_width = scan_event.GetIsolationWidth(0)
            # isolation_center = rawfile.GetPrecursorMassForScanNum(i)
            # isolation_width = rawfile.GetIsolationWidthForScanNum(i)

            if dda:
                # GetMS2MonoMzAndChargeFromScanNum is slow
                mono_mz, charge = _get_mono_and_charge(trailer_data, scan_event)
                if mono_mz <= 0:
                    mono_mz = isolation_center
            else:
                mono_mz = isolation_center
                charge = 0

            # In case that: ms1 = ms_order==2&NCE==0?
            if mono_mz <= 0:
                precursor_mz_values.append(-1.0)
                isolation_mz_lowers.append(-1.0)
                isolation_mz_uppers.append(-1.0)
                precursor_charges.append(0)
                ms_order_list[-1] = 1
            else:
                precursor_mz_values.append(mono_mz)
                isolation_mz_lowers.append(isolation_center - isolation_width / 2)
                isolation_mz_uppers.append(isolation_center + isolation_width / 2)
                precursor_charges.append(charge)
    rawfile.Close()

    # copys of numpy arrays are needed to move them explicitly to cpython heap
    # otherwise mono might interfere later
    spec_dict = {
        "_peak_indices": _peak_indices,
        "peak_mz": np.concatenate(mz_values).copy(),
        "peak_intensity": np.concatenate(intensity_values).copy(),
        "rt": np.array(rt_values).copy(),
        "precursor_mz": np.array(precursor_mz_values).copy(),
        "precursor_charge": np.array(precursor_charges, dtype=np.int8).copy(),
        "isolation_lower_mz": np.array(isolation_mz_lowers).copy(),
        "isolation_upper_mz": np.array(isolation_mz_uppers).copy(),
        "ms_level": np.array(ms_order_list, dtype=np.int8).copy(),
        "nce": np.array(ce_list, dtype=np.float32).copy(),
    }
    for key, val in list(auxiliary_dict.items()):
        auxiliary_dict[key] = np.array(val, dtype=auxiliary_item_dtypes[key]).copy()
    spec_dict.update(auxiliary_dict)
    return spec_dict


def _safe_get_trailer_value(trailer_data, key, default=None, convert_func=None):
    """Safely extract a value from trailer data with optional type conversion.

    Parameters
    ----------
    trailer_data : dict
        Dictionary containing trailer data
    key : str
        Key to extract from trailer data
    default : Any, optional
        Default value if key is not found, by default None
    convert_func : callable, optional
        Function to convert the value (e.g., float, int), by default None

    Returns
    -------
    Any
        Extracted and optionally converted value
    """
    value = trailer_data.get(key, default)
    if value is not None and convert_func is not None:
        try:
            return convert_func(value)
        except (ValueError, TypeError):
            return default
    return value


def _get_mono_and_charge(trailer_data, scan_event):
    mono = float(trailer_data["Monoisotopic M/Z:"].strip())
    charge = int(trailer_data["Charge State:"].strip())
    return mono, charge


def register_readers():
    """
    Register :class:`ThermoRawData` for file formats (types):
    "thermo" and "thermo_raw" in :obj:`alpharaw.ms_data_base.ms_reader_provider`.
    """
    ms_reader_provider.register_reader("thermo", ThermoRawData)
    ms_reader_provider.register_reader("thermo_raw", ThermoRawData)



import numpy as np
from pyteomics import mzml

from .ms_data_base import (
    PEAK_INTENSITY_DTYPE,
    PEAK_MZ_DTYPE,
    MSData_Base,
    ms_reader_provider,
)


class MzMLReader(MSData_Base):
    """
    Load mzml file as `:class:`alpharaw.ms_data_base.MSData_Base` structure.
    This reader will be registered as "mzml" in
    :obj:`alphraw.ms_data_base.ms_reader_provider` by :func:`register_readers`
    """

    def __init__(self, centroided: bool = True, save_as_hdf: bool = False, **kwargs):
        """
        Parameters
        ----------
        centroided : bool, optional
            If peaks will be centroided after loading,
            by default True.

        save_as_hdf : bool, optional
            Automatically save hdf after load raw data, by default False.
        """
        super().__init__(centroided, save_as_hdf, **kwargs)

    def _import(
        self,
        mzml_file_path: str,
    ) -> dict:
        """
        Implementation of :func:`alpharaw.ms_data_base.MSData_Base._import` interface,
        which will be called by :func:`alpharaw.ms_data_base.MSData_Base.import_raw`,
        the main entry of :class:`alpharaw.ms_data_base.MSData_Base` sub-classes.

        Parameters
        ----------
        mzml_file_path : str
            Absolute or relative path of the mzml file.
            For testing purpose, this can be pyteomics `MzML` object as well.

        Returns
        -------
        dict
            Spectrum information dict.
        """
        self.file_type = "mzml"
        if isinstance(mzml_file_path, str):
            reader = mzml.read(mzml_file_path, use_index=True)
        else:
            reader = mzml_file_path
        spec_indices = np.arange(len(reader), dtype=int)

        rt_list = []
        mzs_list = []
        intens_list = []
        ms_level_list = []
        prec_mz_list = []
        charge_list = []
        _peak_indices = []
        isolation_lower_mz_list = []
        isolation_upper_mz_list = []
        nce_list = []

        for _ in spec_indices:
            spec = next(reader)

            (
                rt,
                prec_mz,
                isolation_lower_mz,
                isolation_upper_mz,
                ms_level,
                nce,
                charge,
                masses,
                intensities,
            ) = parse_mzml_entry(spec)

            nce_list.append(nce)

            sortindex = np.argsort(masses)

            masses = masses[sortindex]
            intensities = intensities[sortindex]

            rt_list.append(rt)

            # Remove zero intensities
            to_keep = intensities > 0
            masses = masses[to_keep]
            intensities = intensities[to_keep]

            _peak_indices.append(len(masses))

            mzs_list.append(masses.astype(PEAK_MZ_DTYPE))
            intens_list.append(intensities.astype(PEAK_INTENSITY_DTYPE))
            ms_level_list.append(ms_level)
            prec_mz_list.append(prec_mz)
            charge_list.append(charge)
            isolation_lower_mz_list.append(isolation_lower_mz)
            isolation_upper_mz_list.append(isolation_upper_mz)

        if isinstance(mzml_file_path, str):
            reader.close()

        peak_indices = np.empty(len(spec_indices) + 1, np.int64)
        peak_indices[0] = 0
        peak_indices[1:] = np.cumsum(_peak_indices)
        ret_dict = {
            "peak_indices": peak_indices,
            "peak_mz": np.concatenate(mzs_list),
            "peak_intensity": np.concatenate(intens_list),
            "rt": np.array(rt_list),
            "precursor_mz": np.array(prec_mz_list),
            "precursor_charge": np.array(charge_list, dtype=np.int8),
            "isolation_lower_mz": np.array(isolation_lower_mz_list),
            "isolation_upper_mz": np.array(isolation_upper_mz_list),
            "ms_level": np.array(ms_level_list, dtype=np.int8),
        }
        nce_list = np.array(nce_list, dtype=np.float32)
        if np.any(np.isnan(nce_list)):
            return ret_dict
        ret_dict["nce"] = nce_list
        return ret_dict


def parse_mzml_entry(item_dict: dict) -> tuple:
    """
    Parse mzml entries from pyteomics extracted items.

    Parameters
    ----------
    item_dict : dict
        pyteomics extracted items

    Returns
    -------
    tuple
        items in tuple format.
    """
    rt = float(item_dict.get("scanList").get("scan")[0].get("scan start time"))
    masses = item_dict.get("m/z array")
    intensities = item_dict.get("intensity array")
    ms_level = item_dict.get("ms level")
    prec_mz = -1.0
    isolation_lower_mz = -1.0
    isolation_upper_mz = -1.0
    charge = 0
    nce = 0.0
    if ms_level == 2:
        try:
            charge = int(
                item_dict.get("precursorList")
                .get("precursor")[0]
                .get("selectedIonList")
                .get("selectedIon")[0]
                .get("charge state")
            )
        except TypeError:
            charge = 0
        try:
            charge = int(
                item_dict.get("precursorList")
                .get("precursor")[0]
                .get("selectedIonList")
                .get("selectedIon")[0]
                .get("charge state")
            )
        except TypeError:
            charge = 0

        prec_mz = (
            item_dict.get("precursorList")
            .get("precursor")[0]
            .get("selectedIonList")
            .get("selectedIon")[0]
            .get("selected ion m/z")
        )
        try:
            iso_window = (
                item_dict.get("precursorList")
                .get("precursor")[0]
                .get("isolationWindow")
            )
            iso_lower = float(iso_window.get("isolation window lower offset"))
            iso_upper = float(iso_window.get("isolation window upper offset"))
            isolation_upper_mz = prec_mz + iso_upper
            isolation_lower_mz = prec_mz - iso_lower
        except TypeError:
            isolation_upper_mz = prec_mz + 1.5
            isolation_lower_mz = prec_mz - 1.5

        nce = np.nan
        try:
            filter_string = (
                item_dict.get("scanList").get("scan")[0].get["filter string"]
            )
            if "@hcd" in filter_string:
                nce = float(filter_string.split("@hcd")[1].split(" ")[0])
            elif "@cid" in filter_string:
                nce = float(filter_string.split("@cid")[1].split(" ")[0])
            else:
                nce = np.nan
        except Exception:
            nce = np.nan
    return (
        rt,
        prec_mz,
        isolation_lower_mz,
        isolation_upper_mz,
        ms_level,
        nce,
        charge,
        masses,
        intensities,
    )


def register_readers():
    """
    Register :class:`MzMLReader` for file format "mzml" in
    :obj:`alpharaw.ms_data_base.ms_reader_provider`.
    """
    ms_reader_provider.register_reader("mzml", MzMLReader)



# ruff: noqa: E402  #Module level import not at top of file
import os
import warnings

import numpy as np

from alpharaw.utils.centroiding import naive_centroid

try:
    # require pythonnet, pip install pythonnet on Windows
    import clr

    clr.AddReference("System")

    import System  # noqa: F401
    from System.Globalization import CultureInfo
    from System.Threading import Thread

    from .clr_utils import DotNetArrayToNPArray, ext_dir

    de_fr = CultureInfo("fr-FR")
    other = CultureInfo("en-US")

    Thread.CurrentThread.CurrentCulture = other
    Thread.CurrentThread.CurrentUICulture = other

    clr.AddReference(
        os.path.join(ext_dir, "sciex/Clearcore2.Data.AnalystDataProvider.dll")
    )
    clr.AddReference(os.path.join(ext_dir, "sciex/Clearcore2.Data.dll"))
    clr.AddReference(os.path.join(ext_dir, "sciex/WiffOps4Python.dll"))

    import Clearcore2  # noqa: F401
    import WiffOps4Python  # noqa: F401
    from Clearcore2.Data.AnalystDataProvider import (
        AnalystDataProviderFactory,
        AnalystWiffDataProvider,
    )
    from WiffOps4Python import WiffOps as DotNetWiffOps

    HAS_DOTNET = True
except Exception:
    # allows to use the rest of the code without clr
    warnings.warn(
        "Dotnet-based dependencies could not be loaded. Sciex support is disabled."
    )
    HAS_DOTNET = False


class WillFileReader:
    def __init__(self, filename: str):
        if not HAS_DOTNET:
            raise ValueError(
                "Dotnet-based dependencies are required for reading Sciex files. "
                "Do you have pythonnet and/or mono installed? "
                "See the Readme for details."
            )

        self._wiffDataProvider = AnalystWiffDataProvider()
        self._wiff_file = AnalystDataProviderFactory.CreateBatch(
            filename, self._wiffDataProvider
        )
        self.sample_names = self._wiff_file.GetSampleNames()

    def close(self):
        self._wiffDataProvider.Close()

    def load_sample(
        self,
        sample_id: int,
        centroid: bool = True,
        centroid_ppm: float = 20.0,
        ignore_empty_scans: bool = True,
        keep_k_peaks: int = 2000,
    ):
        if sample_id < 0 or sample_id >= len(self.sample_names):
            raise ValueError("Incorrect sample number.")
        self.wiffSample = self._wiff_file.GetSample(sample_id)
        self.msSample = self.wiffSample.MassSpectrometerSample

        _peak_indices = []
        peak_mz_array_list = []
        peak_intensity_array_list = []
        rt_list = []
        ms_level_list = []
        precursor_mz_list = []
        precursor_charge_list = []
        ce_list = []
        isolation_lower_mz_list = []
        isolation_upper_mz_list = []

        exp_list = [
            self.msSample.GetMSExperiment(i)
            for i in range(self.msSample.ExperimentCount)
        ]

        for j in range(exp_list[0].Details.NumberOfScans):
            for i in range(self.msSample.ExperimentCount):
                exp = exp_list[i]
                massSpectrum = exp.GetMassSpectrum(j)
                massSpectrumInfo = exp.GetMassSpectrumInfo(j)
                details = exp.Details
                ms_level = massSpectrumInfo.MSLevel
                if (
                    ms_level > 1
                    and not details.IsSwath
                    and massSpectrum.NumDataPoints <= 0
                    and ignore_empty_scans
                ):
                    continue
                mz_array = DotNetArrayToNPArray(massSpectrum.GetActualXValues())
                int_array = DotNetArrayToNPArray(
                    massSpectrum.GetActualYValues()
                ).astype(np.float32)
                if centroid:
                    (mz_array, int_array) = naive_centroid(
                        mz_array, int_array, centroiding_ppm=centroid_ppm
                    )
                if len(mz_array) > keep_k_peaks:
                    idxes = np.argsort(int_array)[-keep_k_peaks:]
                    idxes = np.sort(idxes)
                    mz_array = mz_array[idxes]
                    int_array = int_array[idxes]

                peak_mz_array_list.append(mz_array)
                peak_intensity_array_list.append(int_array)

                _peak_indices.append(len(peak_mz_array_list[-1]))
                rt_list.append(exp.GetRTFromExperimentCycle(j))
                # ScanMode = massSpectrumInfo.CentroidMode ? WiffFile.ScanMode.Centroid : WiffFile.ScanMode.Profile,
                # Polarity = (details.Polarity == MSExperimentInfo.PolarityEnum.Positive) ? WiffFile.Polarity.Positive : WiffFile.Polarity.Negative,
                # low_mz = details.StartMass
                # high_mz = details.EndMass
                ms_level_list.append(ms_level)
                # ScanType = (details.IDAType == MSExperimentInfo.IDAExperimentType.Survey) ? WiffFile.ScanType.MS1 : WiffFile.ScanType.MS2,

                center_mz = -1
                isolation_window = 0
                if ms_level > 1:
                    if details.IsSwath and details.MassRangeInfo.Length > 0:
                        center_mz = DotNetWiffOps.get_center_mz(details)
                        isolation_window = DotNetWiffOps.get_isolation_window(details)
                    if isolation_window <= 0:
                        isolation_window = 3.0
                    if center_mz <= 0:
                        center_mz = massSpectrumInfo.ParentMZ
                    precursor_mz_list.append(center_mz)
                    precursor_charge_list.append(massSpectrumInfo.ParentChargeState)
                    ce_list.append(float(massSpectrumInfo.CollisionEnergy))
                    isolation_lower_mz_list.append(center_mz - isolation_window / 2)
                    isolation_upper_mz_list.append(center_mz + isolation_window / 2)
                else:
                    precursor_mz_list.append(-1.0)
                    precursor_charge_list.append(0)
                    ce_list.append(0)
                    isolation_lower_mz_list.append(-1.0)
                    isolation_upper_mz_list.append(-1.0)

        peak_indices = np.empty(len(rt_list) + 1, np.int64)
        peak_indices[0] = 0
        peak_indices[1:] = np.cumsum(_peak_indices)

        return {
            "peak_indices": peak_indices,
            "peak_mz": np.concatenate(peak_mz_array_list),
            "peak_intensity": np.concatenate(peak_intensity_array_list),
            "rt": np.array(rt_list, dtype=np.float64),
            "ms_level": np.array(ms_level_list, dtype=np.int8),
            "precursor_mz": np.array(precursor_mz_list, dtype=np.float64),
            "precursor_charge": np.array(precursor_charge_list, dtype=np.int8),
            "isolation_lower_mz": np.array(isolation_lower_mz_list),
            "isolation_upper_mz": np.array(isolation_upper_mz_list),
            "nce": np.array(ce_list, dtype=np.float32),
        }



# ruff: noqa: E402  #Module level import not at top of file
import os
import warnings

import numpy as np

try:
    import clr

    clr.AddReference("System")

    import ctypes

    from System.Runtime.InteropServices import GCHandle, GCHandleType
except Exception:
    # allows to use the rest of the code without clr
    warnings.warn("Dotnet-based dependencies could not be loaded.")

# from System.Runtime.InteropServices import Marshal
# from System import IntPtr, Int64
# def DotNetArrayToNPArray(src):
#     '''
#     See https://github.com/mobiusklein/ms_deisotope/blob/90b817d4b5ae7823cfe4ad61c57119d62a6e3d9d/ms_deisotope/data_source/thermo_raw_net.py#L217
#     '''
#     if src is None:
#         return np.array([], dtype=np.float64)
#     dest = np.empty(len(src), dtype=np.float64)
#     Marshal.Copy(
#         src, 0,
#         IntPtr.__overloads__[Int64](dest.__array_interface__['data'][0]),
#         len(src))
#     return dest


def DotNetArrayToNPArray(src):
    """
    See https://mail.python.org/pipermail/pythondotnet/2014-May/001527.html
    """
    if src is None:
        return np.array([], dtype=np.float64)
    src_hndl = GCHandle.Alloc(src, GCHandleType.Pinned)
    try:
        src_ptr = src_hndl.AddrOfPinnedObject().ToInt64()
        bufType = ctypes.c_double * len(src)
        cbuf = bufType.from_address(src_ptr)
        dest = np.frombuffer(cbuf, dtype=cbuf._type_).copy()
    finally:
        if src_hndl.IsAllocated:
            src_hndl.Free()
        return dest  # noqa: B012


ext_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), "ext")



import typing
from collections import defaultdict

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from .df_utils import (
    make_psm_plot_df,
    make_query_df_for_peptide,
)
from .plot_utils import plot_scatter

color_map: dict = defaultdict(lambda: "brown")
"""
The colors of different peak annotations. By default:
```
{
    '-': 'lightgrey', # umnatched peaks
    'a': 'darkskyblue',
    'b': 'blue',
    'c': 'skyblue',
    'x': 'darkred',
    'y': 'red',
    'z': 'deeppink',
    'M': 'purple', # precursor
    'Y': 'orange',
    'B': 'darkgreen',
}
```
For other annotations, the default color is "brown".
"""
color_map.update(
    {
        "-": "lightgrey",  # umnatched peaks
        "a": "darkskyblue",
        "b": "blue",
        "c": "skyblue",
        "x": "darkred",
        "y": "red",
        "z": "deeppink",
        "M": "purple",  # precursor
        "Y": "orange",
        "B": "darkgreen",
    }
)


def plot_multi_psms(
    spec_masses_list: typing.List[np.ndarray],
    spec_intens_list: typing.List[np.ndarray],
    sequence: str,
    mods: str,
    mod_sites: str,
    charge: int,
    title: str = "",
    ppm: float = 20.0,
    charged_frag_types: list = ["b_z1", "b_z2", "y_z1", "y_z2"],
    include_fragments: bool = True,
    include_precursor_isotopes: bool = False,
    max_isotope: int = 6,
    min_frag_mz: float = 100.0,
    plot_unmatched_peaks=True,
    match_mode: typing.Literal["closest", "highest"] = "closest",
    plot_template="plotly_white",
    plot_height=600,
    query_left_margin: float = 100000.0,
    query_right_margin: float = 100000.0,
):
    """
    Annotate multiple spectra for a peptide in a single plotly go.Figure.

    Parameters
    ----------
    spec_masses_list : typing.List[np.ndarray]
        The list of peak masses of multiple spectra to be annotated
    spec_intens_list : typing.List[np.ndarray]
        The list of peak intensities of multiple spectra to be annotated
    sequence : str
        Peptide sequence
    mods : str
        Modifications in alphabase format
    mod_sites : str
        Modification sites in alphabase format
    charge : int
        Charge state of the precursor
    title : str, optional
        Plotting title, by default ""
    ppm : float, optional
        Matching mass tolerance in ppm, by default 20.0
    charged_frag_types : list, optional
        Fragment charge states, by default ["b_z1","b_z2","y_z1","y_z2"]
    include_fragments : bool, optional
        If annotate fragments, by default True
    include_precursor_isotopes : bool, optional
        If annotate precursor isotopes, by default False
    max_isotope : int, optional
        Maximal number of isotopes, by default 6
    min_frag_mz : float, optional
        Minimal fragment m/z value to annotate, by default 100.0
    plot_unmatched_peaks : bool, optional
        If plot unmatched peaks (ion_name='-'), by default True
    match_mode : "closest", "highest", optional
        Match the closest or highest peak within the matching tolerance, by default "closest"
    plot_template : str, optional
        Plot template, by default 'plotly_white'
    plot_height : int, optional
        Plot height, by default 600
    query_left_margin : float, optional
        Slice margin to the left mz, by default 100000.0
    query_right_margin : float, optional
        Slice martin to the right mz, by default 100000.0

    Returns
    -------
    go.Figure
        The annotation plots of multiple spectra
    """
    plot_df = make_query_df_for_peptide(
        sequence,
        mods,
        mod_sites,
        charge,
        charged_frag_types=charged_frag_types,
        include_fragments=include_fragments,
        include_precursor_isotopes=include_precursor_isotopes,
        max_isotope=max_isotope,
        min_frag_mz=min_frag_mz,
    )

    slice_masses_list = []
    slice_intens_list = []
    for spec_masses, spec_intens in zip(spec_masses_list, spec_intens_list):
        _slice = (spec_masses >= plot_df.mz.min() - query_left_margin) & (
            spec_masses <= plot_df.mz.max() + query_right_margin
        )
        slice_masses_list.append(spec_masses[_slice])
        slice_intens_list.append(spec_intens[_slice])

    return plot_multi_spectra(
        slice_masses_list,
        slice_intens_list,
        query_masses=plot_df.mz.values,
        query_ion_names=plot_df.ion_name.values,
        query_mass_tols=plot_df.mz.values * ppm * 1e-6,
        title=title,
        plot_unmatched_peaks=plot_unmatched_peaks,
        match_mode=match_mode,
        plot_template=plot_template,
        plot_height=plot_height,
    )


def plot_multi_spectra(
    spec_masses_list: typing.List[np.ndarray],
    spec_intens_list: typing.List[np.ndarray],
    query_masses: np.ndarray,
    query_ion_names: typing.List[str],
    query_mass_tols: np.ndarray,
    title: str = "",
    plot_unmatched_peaks=True,
    match_mode: typing.Literal["closest", "highest"] = "closest",
    plot_template="plotly_white",
    plot_height=600,
):
    """
    Annotate multiple spectra for given queries in a single plotly go.Figure.

    Parameters
    ----------
    spec_masses_list : typing.List[np.ndarray]
        The list of peak masses of multiple spectra to be annotated
    spec_intens_list : typing.List[np.ndarray]
        The list of peak intensities of multiple spectra to be annotated
    query_masses : np.ndarray
        The query m/z values
    query_ion_names : typing.List[str]
        The query ion names
    query_mass_tols : np.ndarray
        The query mass tolerance in Da
    title : str, optional
        The plot title, by default ""
    plot_unmatched_peaks : bool, optional
        If plot unmatched peaks (ion_name='-'), by default True
    match_mode : "closest", "highest", optional
        Match the closest or highest peak within the matching tolerance, by default "closest"
    plot_template : str, optional
        Plot template, by default 'plotly_white'
    plot_height : int, optional
        Plot height, by default 600

    Returns
    -------
    go.Figure
        The annotation plots of multiple spectra
    """
    plot_dfs = []
    for spec_masses, spec_intens in zip(spec_masses_list, spec_intens_list):
        plot_dfs.append(
            make_psm_plot_df(
                spec_masses=spec_masses,
                spec_intensities=spec_intens,
                query_masses=query_masses,
                query_ion_names=query_ion_names,
                query_mass_tols=query_mass_tols,
                query_frag_idxes=np.zeros_like(query_masses, dtype=np.int64),
                modified_sequence="",
                match_mode=match_mode,
            )
        )
    fig = make_subplots(
        rows=len(plot_dfs),
        cols=1,
        shared_xaxes=True,
    )
    layout_vlines = []
    for i in range(len(plot_dfs)):
        _plot = PeakPlot(fig, i + 1)
        _plot.plot(plot_dfs[i], plot_unmatched_peaks=plot_unmatched_peaks)
        layout_vlines.extend(_plot.layout_vlines)

    fig.update_layout(shapes=layout_vlines)

    fig.update_layout(
        template=plot_template,
        title=dict(text=title, yanchor="bottom"),
        hovermode="x",
        height=plot_height,
    )
    fig.update_xaxes(matches="x")
    fig.update_yaxes(
        title="intensity",
    )
    return fig


class PSM_Plot:
    """
    The main class for spectrum annotation of a PSM. It contains three plots:
    1. Ladder plot (`FragCoveragePlot`) for fragment coverage of the peptide.
    2. Peak annotation plot (`PeakPlot`) for the spectrum.
    3. Matching mass error plot (`MassErrPlot`) for the matched peaks.
    """

    vertical_spacing = 0.05
    template = "plotly_white"
    plot_height = 600

    def __init__(
        self,
        peak_plot_rows: int = 4,
        mass_err_plot_rows: int = 1,
        frag_coverage_plot_rows: int = 1,
        frag_coverage: bool = True,
    ):
        """
        Parameters
        ----------
        peak_plot_rows : int, optional
            The height (ratio) of peak plot, by default 4
        mass_err_plot_rows : int, optional
            The height (ratio) of mass error plot, by default 1
        frag_coverage_plot_rows : int, optional
            The height (ratio) of fragment coverage plot, by default 1
        frag_coverage : bool, optional
            If plot fragment coverage, by default True
        """
        specs = []
        if frag_coverage:
            specs.append(
                [{"rowspan": frag_coverage_plot_rows, "colspan": 3}, None, None]
            )
            specs.extend([[None, None, None]] * (frag_coverage_plot_rows - 1))
        specs.append([{"rowspan": peak_plot_rows, "colspan": 3}, None, None])
        specs.extend([[None, None, None]] * (peak_plot_rows - 1))
        specs.append([{"rowspan": mass_err_plot_rows, "colspan": 3}, None, None])
        specs.extend([[None, None, None]] * (mass_err_plot_rows - 1))
        if not frag_coverage:
            specs.append(
                [{"rowspan": frag_coverage_plot_rows, "colspan": 3}, None, None]
            )
            specs.extend([[None, None, None]] * (frag_coverage_plot_rows - 1))

        if frag_coverage:
            (
                frag_cov_row,
                peak_row,
                mass_err_row,
            ) = np.cumsum(
                [
                    1,
                    frag_coverage_plot_rows,
                    peak_plot_rows,
                ]
            )
        else:
            (
                peak_row,
                mass_err_row,
                frag_cov_row,
            ) = np.cumsum([1, peak_plot_rows, mass_err_plot_rows])

        self.specs = specs
        self.peak_row = peak_row
        self.mass_err_row = mass_err_row
        self.frag_cov_row = frag_cov_row
        self.rows = peak_plot_rows + mass_err_plot_rows + frag_coverage_plot_rows

    def plot(
        self,
        plot_df: pd.DataFrame,
        sequence: str,
        title: str,
        plot_unmatched_peaks: bool = False,
    ) -> go.Figure:
        """
        Main entry of `PSM_Plot` for peak annotation.

        Parameters
        ----------
        plot_df : pd.DataFrame
            The plot_df can be generated by
            :func:`alpharaw.viz.df_utils.make_psm_plot_df_for_peptide`,
            :func:`alpharaw.viz.df_utils.make_psm_plot_for_frag_dfs`, or
            :func:`alpharaw.viz.df_utils.make_psm_plot_df`.
        sequence : str
            Peptide sequence, for fragment coverage plot
        title : str
            Plot title
        plot_unmatched_peaks : bool, optional
            If plot unmatched peaks with ion_name `-`, by default False

        Returns
        -------
        go.Figure
            Peak annotation plot in plotly go.Figure
        """
        if "pcc" in plot_df.columns and len(plot_df) > 0:
            title = f"{title} (R={plot_df.pcc.values[0]:.3f})"
        self._init_plot(title)

        self.peak_plot.plot(plot_df, plot_unmatched_peaks=plot_unmatched_peaks)
        self.mass_err_plot.plot(
            plot_df,
        )
        self.frag_cov_plot.plot(plot_df, sequence)

        self.fig.update_layout(shapes=self.peak_plot.layout_vlines)

        return self.fig

    def _init_plot(self, title):
        self.fig = make_subplots(
            rows=self.rows,
            cols=3,
            shared_xaxes=True,
            specs=self.specs,
            vertical_spacing=self.vertical_spacing,
            column_widths=[0.25, 0.5, 0.25],
        )

        self.peak_plot = PeakPlot(
            self.fig,
            self.peak_row,
        )
        self.mass_err_plot = MassErrPlot(
            self.fig,
            self.mass_err_row,
        )
        self.frag_cov_plot = FragCoveragePlot(self.fig, self.frag_cov_row)

        self.fig.update_layout(
            template=self.template,
            title=dict(text=title, yanchor="bottom"),
            hovermode="x",
            height=self.plot_height,
        )
        self.fig.update_xaxes(matches="x")
        self.fig.update_yaxes(
            title="intensity",
        )


class MassErrPlot:
    def __init__(self, fig_subplots, row):
        self.fig = fig_subplots
        self.row = row
        self.col = 1
        self.hovertemplate = "%{hovertext}<br>" "<b>m/z:</b> %{x}<br><b>ppm:</b> %{y}"

    def plot(self, plot_df: pd.DataFrame):
        if "color" not in plot_df.columns:
            plot_df["color"] = [
                color_map[ion_type] for ion_type in plot_df.ion_name.str[0].values
            ]

        for color, df in plot_df.query("intensity>0 and ion_name!='-'").groupby(
            "color"
        ):
            self._plot_one_type(df, color)

        self.fig.update_yaxes(title_text="ppm", row=self.row, col=self.col)
        return self.fig

    def _plot_one_type(self, df, color):
        self.fig.add_trace(
            plot_scatter(
                df.mz.values,
                df.ppm_err.round(4).values,
                color=color,
                marker_size=5,
                hovertext=df.ion_name.values,
                hovertemplate=self.hovertemplate,
                name=color,
            ),
            row=self.row,
            col=self.col,
        )


class PeakPlot:
    fig: go.Figure
    row: int = 1
    col: int = 1
    hovertemplate = (
        "<b>%{hovertext}</b><br>" "<b>m/z:</b> %{x}<br>" "<b>intensity:</b> %{y}"
    )
    peak_line_width = 1.5

    def __init__(self, fig: go.Figure, row: int, col: int = 1):
        self.fig = fig
        self.row = row
        self.col = col

    def plot(
        self,
        plot_df: pd.DataFrame,
        plot_unmatched_peaks: bool = True,
    ) -> go.Figure:
        if "color" not in plot_df.columns:
            plot_df["color"] = [
                color_map[ion_type] for ion_type in plot_df.ion_name.str[0].values
            ]
        plot_df.loc[plot_df.ion_name == "-", "color"] = color_map["-"]
        if plot_unmatched_peaks:
            _df = plot_df[plot_df.ion_name == "-"]
            self.fig.add_trace(
                plot_scatter(
                    _df.mz,
                    _df.intensity,
                    color=color_map["-"],
                    marker_size=1,
                    hovertext=None,
                    hovertemplate=None,
                    name="",
                ),
                row=self.row,
                col=self.col,
            )
        else:
            plot_df = plot_df.query('ion_name != "-"')

        matched_df = plot_df.query('ion_name != "-"')
        for color, df in matched_df.groupby("color"):
            self._plot_one_ion_type_scatter(df, color)

        self._get_peak_vlines(
            plot_df,
        )

        self._plot_ion_name(matched_df)

        return self.fig

    def _plot_one_ion_type_scatter(self, df, color):
        _df = df.query("intensity>0")
        self.fig.add_trace(
            plot_scatter(
                _df.mz,
                _df.intensity,
                color=color,
                marker_size=1,
                hovertext=_df.ion_name,
                hovertemplate=self.hovertemplate,
                name="",
            ),
            row=self.row,
            col=self.col,
        )
        _df = df.query("intensity<0")
        self.fig.add_trace(
            plot_scatter(
                _df.mz,
                _df.intensity,
                color=color,
                marker_size=1,
                hovertext=_df.ion_name,
                hovertemplate=self.hovertemplate,
                name="",
            ),
            row=self.row,
            col=self.col,
        )

    def _plot_ion_name(self, plot_df):
        df = plot_df.query("intensity>0")
        max_inten = df.intensity.max()
        yshift = max_inten * 0.02
        for mz, inten, ion in df[["mz", "intensity", "ion_name"]].values:
            self.fig.add_annotation(
                x=mz,
                y=inten + yshift,
                text=ion,
                textangle=-90,
                font_size=10,
                row=self.row,
                col=self.col,
            )

        pred_df = plot_df.query("intensity<0")
        pred_df = pred_df[~pred_df.ion_name.isin(set(df.ion_name))]
        for mz, inten, ion in pred_df[["mz", "intensity", "ion_name"]].values:
            self.fig.add_annotation(
                x=mz,
                y=inten - yshift,
                text=ion,
                textangle=-90,
                font_size=10,
                ay=inten - yshift - max_inten * (0.28 + len(ion) / 60),
                ayref=f"y{self.row}",
                yref=f"y{self.row}",
                row=self.row,
                col=self.col,
            )

    def _get_peak_vlines(
        self,
        plot_df,
    ):
        self.layout_vlines = [
            dict(
                type="line",
                xref=f"x{self.row}",
                yref=f"y{self.row}",
                x0=plot_df.loc[i, "mz"],
                y0=0,
                x1=plot_df.loc[i, "mz"],
                y1=plot_df.loc[i, "intensity"],
                line=dict(
                    color=plot_df.loc[i, "color"],
                    width=self.peak_line_width,
                ),
            )
            for i in plot_df.index
        ]
        # for i in plot_df.index:
        #     self.fig.add_shape(type='line',
        #         x0=plot_df.loc[i, 'mz'],
        #         x1=plot_df.loc[i, 'mz'],
        #         y0=0,
        #         y1=plot_df.loc[i, 'intensity'],
        #         line_color = color_map[plot_df.loc[i, 'ion_name'][0]],
        #         line_width = self.peak_line_width,
        #         row=self.row,
        #         col=self.col,
        #     )


class FragCoveragePlot:
    def __init__(self, fig_subplots, row):
        self.fig = fig_subplots
        self.row = row
        self.col = 1
        self.font_size_sequence = 14
        self.font_size_coverage = 8
        self.mod_aa_color = "firebrick"

    def plot(
        self,
        plot_df,
        sequence,
    ):
        if "color" not in plot_df.columns:
            plot_df["color"] = [
                color_map[ion_type] for ion_type in plot_df.ion_name.str[0].values
            ]
        if len(sequence) > 0:
            d = (plot_df.mz.max() - plot_df.mz.min()) * 2 / len(sequence)
            aa_x_position_name = np.linspace(
                plot_df.mz.min() + d, plot_df.mz.max() - d, len(sequence) + 1
            )

            colors = [None] * len(sequence)
            if "mod_sites" in plot_df.columns:
                mod_sites = plot_df.mod_sites.values[0]
                if len(mod_sites) > 0:
                    for site in mod_sites.split(";"):
                        site = int(site)
                        if site == 0 or site == -1:
                            colors[site] = self.mod_aa_color
                        else:
                            colors[site - 1] = self.mod_aa_color

            self._plot_sequence(sequence, aa_x_position_name, colors)
            for ion_type in np.unique(plot_df.ion_name.str[0]):
                self._plot_coverage_one_frag_type(
                    plot_df,
                    sequence,
                    aa_x_position_name,
                    ion_type,
                )
            # self._plot_coverage_one_frag_type(
            #     plot_df, sequence, aa_x_position_name,
            #     'b', color_map['b'],
            # )
            # self._plot_coverage_one_frag_type(
            #     plot_df, sequence, aa_x_position_name,
            #     'y', color_map['y'],
            # )
            self.fig.update_yaxes(
                visible=False,
                range=(-1.1, 1.1),
                row=self.row,
                col=self.col,
            )
            self.fig.update_xaxes(
                visible=False,
                row=self.row,
                col=self.col,
            )
        return self.fig

    def _plot_sequence(
        self,
        sequence,
        aa_x_position_name,
        colors,
    ):
        for i, aa in enumerate(sequence):
            self.fig.add_annotation(
                dict(
                    text=aa,
                    x=aa_x_position_name[i],
                    y=0,
                    showarrow=False,
                    font=dict(
                        size=self.font_size_sequence,
                        color=colors[i],
                    ),
                    yshift=1,
                    align="center",
                ),
                row=self.row,
                col=self.col,
            )

    def _plot_coverage_one_frag_type(
        self,
        plot_df,
        sequence,
        aa_x_position_name,
        ion_type,
    ):
        nAA = len(sequence)
        plot_df = plot_df[plot_df.ion_name.str.startswith(ion_type)].query(
            "intensity>0"
        )

        covs = np.zeros(
            max(plot_df.fragment_site.max() + 2 if len(plot_df) > 0 else 0, nAA),
            dtype=np.int64,
        )
        cov_colors = [""] * len(covs)
        if ion_type in "abc":
            for frag_idx, color in zip(
                plot_df.fragment_site.values, plot_df.color.values
            ):
                covs[frag_idx] = 1
                cov_colors[frag_idx] = color
        elif ion_type in "xyz":
            for frag_idx, color in zip(
                plot_df.fragment_site.values, plot_df.color.values
            ):
                covs[frag_idx + 1] = 1
                cov_colors[frag_idx + 1] = color

        def get_position_name(ion_type, i):
            if ion_type in "abc":
                return dict(
                    x=[
                        aa_x_position_name[i],
                        aa_x_position_name[i]
                        + (aa_x_position_name[i + 1] - aa_x_position_name[i]) / 2,
                        aa_x_position_name[i]
                        + (aa_x_position_name[i + 1] - aa_x_position_name[i]) / 2,
                    ],
                    y=[0.6, 0.6, 0.4],
                    tx=aa_x_position_name[i]
                    + (aa_x_position_name[i + 1] - aa_x_position_name[i]) / 4,
                    ty=1.0,
                    s=i + 1,
                )
            elif ion_type in "xyz":
                return dict(
                    x=[
                        aa_x_position_name[i],
                        aa_x_position_name[i]
                        - (aa_x_position_name[i + 1] - aa_x_position_name[i]) / 2,
                        aa_x_position_name[i]
                        - (aa_x_position_name[i + 1] - aa_x_position_name[i]) / 2,
                    ],
                    y=[-0.6, -0.6, -0.4],
                    tx=aa_x_position_name[i]
                    - (aa_x_position_name[i + 1] - aa_x_position_name[i]) / 4,
                    ty=-1.0,
                    s=nAA - i,
                )

        for i, (cov, color) in enumerate(zip(covs[:nAA], cov_colors[:nAA])):
            if cov:
                pos = get_position_name(ion_type, i)
                self.fig.add_trace(
                    go.Scatter(
                        x=pos["x"],
                        y=pos["y"],
                        mode="lines",
                        showlegend=False,
                        marker_color=color,
                        line_width=1,
                        hoverinfo="skip",
                    ),
                    row=self.row,
                    col=self.col,
                )
                self.fig.add_annotation(
                    dict(
                        text=f"{ion_type}{pos['s']}",
                        x=pos["tx"],
                        y=pos["ty"],
                        showarrow=False,
                        font_size=self.font_size_coverage,
                    ),
                    row=self.row,
                    col=self.col,
                )



from typing import Tuple

import numba
import numpy as np


@numba.njit
def match_batch_spec(
    spec_idxes: np.ndarray,
    peak_mzs: np.ndarray,
    peak_intens: np.ndarray,
    peak_start_idxes: np.ndarray,
    peak_stop_idxes: np.ndarray,
    query_mzs: np.ndarray,
    query_mz_tols: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Extract matched mzs and intensities for query m/z values against the given batch spectra.

    Parameters
    ----------
    spec_idxes : np.ndarray
        The batch spectra, given as spectrum indexes.
    peak_mzs : np.ndarray
        The peak m/z values in the whole raw data.
    peak_intens : np.ndarray
        The peak intensities in the whole raw data.
    peak_start_idxes : np.ndarray
        The batch spectra, given as the start indexes in peak m/z and intensities.
    peak_stop_idxes : np.ndarray
        The batch spectra, given as the stop indexes in peak m/z and intensities.
    query_mzs : np.ndarray
        The query m/z values, these can be from fragments of a precursor.
    query_mz_tols : np.ndarray
        The query tolerance values of query_mzs.

    Returns
    -------
    Tuple[ndarray, ndarray]
        ndarray with shape (spectrum num, query num): matched m/z values. 0.0 if not matched.
        ndarray with shape (spectrum num, query num): matched intensity values. 0.0 if not matched.
    """
    matched_mzs = np.zeros((len(spec_idxes), len(query_mzs)), dtype=peak_mzs.dtype)
    matched_intens = np.zeros(
        (len(spec_idxes), len(query_mzs)), dtype=peak_intens.dtype
    )

    for i_spec, spec_idx in enumerate(spec_idxes):
        cur_peak_mzs = peak_mzs[peak_start_idxes[spec_idx] : peak_stop_idxes[spec_idx]]
        cur_peak_intens = peak_intens[
            peak_start_idxes[spec_idx] : peak_stop_idxes[spec_idx]
        ]

        idxes = np.searchsorted(cur_peak_mzs, query_mzs)

        for i, idx in enumerate(idxes):
            if idx == 0:
                if abs(query_mzs[i] - cur_peak_mzs[idx]) <= query_mz_tols[i]:
                    matched_mzs[i_spec, i] = cur_peak_mzs[idx]
                    matched_intens[i_spec, i] = cur_peak_intens[idx]
            elif idx == len(cur_peak_mzs):
                if abs(query_mzs[i] - cur_peak_mzs[idx - 1]) <= query_mz_tols[i]:
                    matched_mzs[i_spec, i] = cur_peak_mzs[idx - 1]
                    matched_intens[i_spec, i] = cur_peak_intens[idx - 1]
            else:
                left_dist = abs(query_mzs[i] - cur_peak_mzs[idx - 1])
                right_dist = abs(query_mzs[i] - cur_peak_mzs[idx])
                if right_dist <= query_mz_tols[i] and left_dist <= query_mz_tols[i]:
                    matched_mzs[i_spec, i] = (
                        cur_peak_mzs[idx] * cur_peak_intens[idx]
                        + cur_peak_mzs[idx - 1] * cur_peak_intens[idx - 1]
                    ) / (cur_peak_intens[idx] + cur_peak_intens[idx - 1])
                    matched_intens[i_spec, i] = (
                        cur_peak_intens[idx] + cur_peak_intens[idx - 1]
                    )
                elif left_dist <= query_mz_tols[i]:
                    matched_mzs[i_spec, i] = cur_peak_mzs[idx - 1]
                    matched_intens[i_spec, i] = cur_peak_intens[idx - 1]
                elif right_dist <= query_mz_tols[i]:
                    matched_mzs[i_spec, i] = cur_peak_mzs[idx]
                    matched_intens[i_spec, i] = cur_peak_intens[idx]
    return matched_mzs, matched_intens


@numba.njit
def match_closest_peaks(
    spec_mzs: np.ndarray,
    spec_intens: np.ndarray,
    query_mzs: np.ndarray,
    query_mz_tols: np.ndarray,
) -> np.ndarray:
    """Matching query mz values against sorted MS2/spec m/z values,
    only closest (minimal abs mass error) peaks are returned.

    Parameters
    ----------
    spec_mzs : np.ndarray
        sorted 1-D mz array of the spectrum

    spec_intens : np.ndarray
        1-D intensity array of the spectrum. Not used here

    query_mzs : np.ndarray
        query n-D mz array

    query_mz_tols : np.ndarray
        query n-D mz tolerance array, same shape as query_mzs

    Returns
    -------
    np.ndarray
        Matched indices of spec_mzs, -1 means no peaks were matched.
        Same shape as query_mzs
    """
    mzs = query_mzs.reshape(-1)
    query_mz_tols = query_mz_tols.reshape(-1)
    query_left_mzs = mzs - query_mz_tols
    query_right_mzs = mzs + query_mz_tols
    idxes = np.searchsorted(spec_mzs, query_left_mzs)
    ret_indices = np.empty_like(mzs, dtype=np.int32)
    for i, idx in enumerate(idxes):
        min_merr = 1000000
        closest_idx = -1
        for _idx in range(idx, len(spec_mzs)):
            if spec_mzs[_idx] > query_right_mzs[i]:
                break
            elif spec_mzs[_idx] < query_left_mzs[i]:
                continue
            elif min_merr > abs(spec_mzs[_idx] - mzs[i]):
                min_merr = abs(spec_mzs[_idx] - mzs[i])
                closest_idx = _idx
        ret_indices[i] = closest_idx
    return ret_indices.reshape(query_mzs.shape)


@numba.njit
def match_highest_peaks(
    spec_mzs: np.ndarray,
    spec_intens: np.ndarray,
    query_mzs: np.ndarray,
    query_mz_tols: np.ndarray,
) -> np.ndarray:
    """Matching query mz values against sorted MS2/spec masses,
    only highest peaks are returned.

    Parameters
    ----------
    spec_mzs : np.ndarray
        sorted 1-D mz array of the spectrum

    spec_intens : np.ndarray
        1-D intensity array of the spectrum. Not used here

    query_mzs : np.ndarray
        query n-D mz array

    query_mz_tols : np.ndarray
        query n-D mz tolerance array, same shape as query_mzs

    Returns
    -------
    np.ndarray
        Matched indices of spec_mzs, -1 means no peaks were matched.
        Same shape as query_mzs
    """
    mzs = query_mzs.reshape(-1)
    query_mz_tols = query_mz_tols.reshape(-1)
    query_left_mzs = mzs - query_mz_tols
    query_right_mzs = mzs + query_mz_tols
    idxes = np.searchsorted(spec_mzs, query_left_mzs)
    ret_indices = np.empty_like(mzs, dtype=np.int32)
    for i, idx in enumerate(idxes):
        highest = 0
        highest_idx = -1
        for _idx in range(idx, len(spec_mzs)):
            if spec_mzs[_idx] > query_right_mzs[i]:
                break
            elif spec_mzs[_idx] < query_left_mzs[i]:
                continue
            elif highest < spec_intens[_idx]:
                highest = spec_intens[_idx]
                highest_idx = _idx
        ret_indices[i] = highest_idx
    return ret_indices.reshape(query_mzs.shape)


@numba.njit
def match_profile_peaks(
    spec_mzs: np.ndarray,
    spec_intens: np.ndarray,
    query_mzs: np.ndarray,
    query_mz_tols: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Matching query mz values against sorted MS2/spec profile masses,
    both left- and right-most m/z values are returned.

    Parameters
    ----------
    spec_mzs : np.ndarray
        sorted 1-D mz array of the spectrum

    spec_intens : np.ndarray
        1-D intensity array of the spectrum. Not used here

    query_mzs : np.ndarray
        query n-D mz array

    query_mz_tols : np.ndarray
        query n-D mz tolerance array, same shape as query_mzs

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        np.ndarray: matched first (left-most) indices, the shape is the same as query_mzs.
        -1 means no peaks are matched for the query mz.

        np.ndarray: matched last (right-most) indices, the shape is the same as query_mzs.
        -1 means no peaks are matched for the query mz.
    """
    mzs = query_mzs.reshape(-1)
    query_mz_tols = query_mz_tols.reshape(-1)
    query_left_mzs = mzs - query_mz_tols
    query_right_mzs = mzs + query_mz_tols
    idxes = np.searchsorted(spec_mzs, query_left_mzs)
    match_start_indices = np.full_like(mzs, -1, dtype=np.int32)
    match_stop_indices = np.full_like(mzs, -1, dtype=np.int32)
    for i, idx in enumerate(idxes):
        for first_idx in range(idx, len(spec_mzs)):
            if spec_mzs[first_idx] < query_left_mzs[i]:
                continue
            elif spec_mzs[first_idx] > query_right_mzs[i]:
                break
            else:
                match_start_indices[i] = first_idx
                for last_idx in np.arange(first_idx + 1, len(spec_mzs)):
                    if spec_mzs[last_idx] > query_right_mzs[i]:
                        match_stop_indices[i] = last_idx
                        break
                if match_stop_indices[i] == -1:
                    match_stop_indices[i] = len(spec_mzs)
                break
    return (
        match_start_indices.reshape(query_mzs.shape),
        match_stop_indices.reshape(query_mzs.shape),
    )



from typing import Tuple, Union

import numba
import numpy as np
import pandas as pd
import tqdm
from alphabase.peptide.fragment import (
    create_fragment_mz_dataframe,
    get_charged_frag_types,
)

from alpharaw import register_all_readers
from alpharaw.dia.normal_dia import NormalDIAGrouper
from alpharaw.match.match_utils import (
    match_closest_peaks,
    match_highest_peaks,
)
from alpharaw.match.spec_finder import find_dia_spec_idxes_same_window
from alpharaw.ms_data_base import (
    PEAK_INTENSITY_DTYPE,
    PEAK_MZ_DTYPE,
    MSData_Base,
    ms_reader_provider,
)
from alpharaw.utils.ms_path_utils import parse_ms_files_to_dict

register_all_readers()  # TODO remove this import side effect


class PepSpecMatch:
    """
    Extract fragment ions from MS2 data.
    The extracted information can be used for visualization of peak annotation or
    PeptDeep transfer learnining for the MS2 model.
    """

    match_closest: bool = True
    use_ppm: bool = True
    #: matching mass tolerance
    tolerance: float = 20.0
    ms_loader_thread_num: int = 4

    def __init__(
        self,
        charged_frag_types: list = None,
        match_closest: bool = True,
        use_ppm: bool = True,
        tol_value: float = 20.0,
    ):
        """
        Parameters
        ----------
        charged_frag_types : list, optional
            fragment types with charge states,
            e.g. ['b_z1', 'y_z2', 'b_modloss_z1', 'y_H2O_z2'].
            Defaults to `get_charged_frag_types(['b','y','b_modloss','y_modloss'], 2)`.

        match_closest : bool, optional
            if True, match the closest peak for a m/z;
            if False, matched the higest peak for a m/z in the tolerance range.
            By default True.

        use_ppm : bool, optional
            If use ppm other wise Da, by default True.

        tol_value : float, optional
            Matching tolerance value (ppm or Da based on `use_ppm`)
            for peak annotation, by default 20.0
        """
        self.charged_frag_types = (
            get_charged_frag_types(["b", "y", "b_modloss", "y_modloss"], 2)
            if charged_frag_types is None
            else charged_frag_types
        )
        self.match_closest = match_closest
        self.use_ppm = use_ppm
        self.tolerance = tol_value

    def get_fragment_mz_df(self) -> pd.DataFrame:
        """
        Call :func:`alphabase.peptide.fragment.create_fragment_mz_dataframe`
        for :attr:`PepSpecMatch.psm_df` and :attr:`PepSpecMatch.charged_frag_types`.


        Returns
        -------
        DataFrame
            The fragment m/z dataframe in alphabase format.
        """
        return create_fragment_mz_dataframe(
            self.psm_df,
            self.charged_frag_types,
            dtype=PEAK_MZ_DTYPE,
        )

    def _add_missing_columns_to_psm_df(
        self, psm_df: pd.DataFrame, raw_data: MSData_Base = None
    ):
        """
        Add missing "rt", "nce", "rt_norm", ("mobility") columns to `psm_df` inplace if missing.

        Parameters
        ----------
        psm_df : pd.DataFrame
            psm dataframe to be processed.
        raw_data : MSData_Base, optional
            The `MSData_Base`. If None, `self.raw_data`. by default None.

        Returns
        -------
        DataFrame
            The original `psm_df` with missing columns added.
        """
        if raw_data is None:
            raw_data = self.raw_data
        add_spec_info_list = []
        if "rt" not in psm_df.columns:
            add_spec_info_list.append("rt")

        if "nce" in raw_data.spectrum_df.columns and "nce" not in psm_df.columns:
            add_spec_info_list.append("nce")

        if (
            "mobility" not in psm_df.columns
            and "mobility" in raw_data.spectrum_df.columns
        ):
            add_spec_info_list.append("mobility")

        if len(add_spec_info_list) > 0:
            # pfind does not report RT in the result file
            psm_df = (
                psm_df.reset_index()
                .merge(
                    raw_data.spectrum_df[["spec_idx"] + add_spec_info_list],
                    how="left",
                    on="spec_idx",
                )
                .set_index("index")
            )

            if "rt" in add_spec_info_list:
                psm_df["rt_norm"] = psm_df.rt / raw_data.spectrum_df.rt.max()
        # if 'rt_sec' not in psm_df.columns:
        #     psm_df['rt_sec'] = psm_df.rt*60
        return psm_df

    def _prepare_matching_dfs(self) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """
        Prepare empty `fragment_mz_df`, `matched_intensity_df`,
        and `matched_mz_err_df` dataframes to extract peak matching information
        for `self.psm_df`. These three dataframes will be only used internally
        in :class:`PepSpecMatch` objects.

        Returns
        -------
        Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]
            pd.DataFrame: fragment mz dataframe.

            pd.DataFrame: intensity dataframe to match.

            pd.DataFrame: mz error dataframe to match.
        """
        fragment_mz_df = self.get_fragment_mz_df()

        matched_intensity_df = pd.DataFrame(
            np.zeros_like(fragment_mz_df.values, dtype=PEAK_INTENSITY_DTYPE),
            columns=fragment_mz_df.columns,
        )

        matched_mz_err_df = pd.DataFrame(
            np.full_like(fragment_mz_df.values, np.inf, dtype=PEAK_MZ_DTYPE),
            columns=fragment_mz_df.columns,
        )
        return (fragment_mz_df, matched_intensity_df, matched_mz_err_df)

    def load_ms_data(
        self,
        ms_file: Union[str, MSData_Base],
        ms_file_type: str = "alpharaw_hdf",
        process_count: int = 8,
        **kwargs,
    ):
        """Load MS file to set `self.raw_data`.

        Parameters
        ----------
        ms_file : str | MSData_Base
            Absolute or relative path of the ms2 file.

        ms_file_type : str, optional
            ms2 file type, could be
            ["alpharaw_hdf","thermo","sciex","alphapept_hdf","mgf"].
            Default to 'alpharaw_hdf'.
        """
        self.raw_data = load_ms_data(ms_file, ms_file_type, process_count=process_count)

    def get_peaks(self, spec_idx: int, **kwargs):
        return self.raw_data.get_peaks(spec_idx)

    def _match_one_psm(
        self,
        peak_mzs: np.ndarray,
        peak_intensities: np.ndarray,
        fragment_mz_df: pd.DataFrame,
        matched_intensity_df: pd.DataFrame,
        matched_mz_err_df: pd.DataFrame,
        frag_start_idx: int,
        frag_stop_idx: int,
    ):
        """
        Match fragments of one precursor (located by `frag_start_idx` and `frag_stop_idx`)
        against the corresponding `peak_mzs`.

        Parameters
        ----------
        peak_mzs : np.ndarray
            Peak m/z values to be matched.
        peak_intensities : np.ndarray
            Peak intensities to be matched.
        fragment_mz_df : pd.DataFrame
            fragment m/z dataframe to be matched.
        matched_intensity_df : pd.DataFrame
            The dataframe to store matched intensity values.
        matched_mz_err_df : pd.DataFrame
            The dataframe to store matched mz error values.
        frag_start_idx : int
            fragment start index of the given PSM.
        frag_stop_idx : int
            fragment stop index of the given PSM.
        """
        if len(peak_mzs) == 0:
            return

        peak_mzs = peak_mzs.astype(PEAK_MZ_DTYPE)

        frag_mzs = fragment_mz_df.values[frag_start_idx:frag_stop_idx, :]

        if self.use_ppm:
            mz_tols = frag_mzs * self.tolerance * 1e-6
        else:
            mz_tols = np.full_like(frag_mzs, self.tolerance)

        if self.match_closest:
            matched_idxes = match_closest_peaks(
                peak_mzs, peak_intensities, frag_mzs, mz_tols
            )
        else:
            matched_idxes = match_highest_peaks(
                peak_mzs,
                peak_intensities,
                frag_mzs,
                mz_tols,
            )

        matched_intens = peak_intensities[matched_idxes]
        matched_intens[matched_idxes == -1] = 0

        matched_mz_errs = np.abs(peak_mzs[matched_idxes] - frag_mzs)
        matched_mz_errs[matched_idxes == -1] = np.inf

        matched_intensity_df.values[frag_start_idx:frag_stop_idx, :] = matched_intens

        matched_mz_err_df.values[frag_start_idx:frag_stop_idx, :] = matched_mz_errs

    def match_ms2_one_raw(
        self,
        psm_df_one_raw: pd.DataFrame,
        verbose: bool = False,
    ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """
        Matching psm_df_one_raw against self.raw_data
        after `self.load_ms_data()`

        Parameters
        ----------
        psm_df_one_raw : pd.DataFrame
            psm dataframe
            that contains only one raw file

        Returns
        -------
        Tuple:
            pd.DataFrame: psm dataframe with fragment index information.

            pd.DataFrame: fragment mz dataframe.

            pd.DataFrame: matched intensity dataframe.

            pd.DataFrame: matched mass error dataframe.
              np.inf if a fragment is not matched.
        """
        self.psm_df = psm_df_one_raw

        psm_df_one_raw = self._add_missing_columns_to_psm_df(
            psm_df_one_raw, self.raw_data
        )

        (
            fragment_mz_df,
            matched_intensity_df,
            matched_mz_err_df,
        ) = self._prepare_matching_dfs()

        psm_iters = psm_df_one_raw[
            ["spec_idx", "frag_start_idx", "frag_stop_idx"]
        ].values
        if verbose:
            psm_iters = tqdm.tqdm(psm_iters)

        for spec_idx, frag_start_idx, frag_stop_idx in psm_iters:
            (spec_mzs, spec_intens) = self.get_peaks(spec_idx)

            self._match_one_psm(
                spec_mzs,
                spec_intens,
                fragment_mz_df,
                matched_intensity_df,
                matched_mz_err_df,
                frag_start_idx,
                frag_stop_idx,
            )

        return (psm_df_one_raw, fragment_mz_df, matched_intensity_df, matched_mz_err_df)

    def _match_ms2_one_raw_numba(self, raw_name: str, psm_df_one_raw: pd.DataFrame):
        if raw_name in self._ms_file_dict:
            raw_data = load_ms_data(
                self._ms_file_dict[raw_name],
                self._ms_file_type,
                process_count=self.ms_loader_thread_num,
            )
            if raw_data is None:
                return

            psm_df_one_raw = self._add_missing_columns_to_psm_df(
                psm_df_one_raw, raw_data
            )

            if self.use_ppm:
                all_frag_mz_tols = self.fragment_mz_df.values * self.tolerance * 1e-6
            else:
                all_frag_mz_tols = np.full_like(
                    self.fragment_mz_df.values, self.tolerance
                )

            match_one_raw_with_numba(
                psm_df_one_raw.spec_idx.values,
                psm_df_one_raw.frag_start_idx.values,
                psm_df_one_raw.frag_stop_idx.values,
                self.fragment_mz_df.values,
                all_frag_mz_tols,
                raw_data.peak_df.mz.values,
                raw_data.peak_df.intensity.values,
                raw_data.spectrum_df.peak_start_idx.values,
                raw_data.spectrum_df.peak_stop_idx.values,
                self.matched_intensity_df.values,
                self.matched_mz_err_df.values,
                self.match_closest,
            )
        else:
            print(f"`{raw_name}` is not found in ms_file_dict.")
            return
        return psm_df_one_raw

    def match_ms2_multi_raw(
        self,
        psm_df: pd.DataFrame,
        ms_files: Union[dict, list],
        ms_file_type: str = "alpharaw_hdf",
        process_num: int = 1,
    ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """
        Matching PSM dataframe against the ms2 files in ms_files
        This method will store matched values as attributes:
        - self.psm_df
        - self.fragment_mz_df
        - self.matched_intensity_df
        - self.matched_mz_err_df

        Parameters
        ----------
        psm_df : pd.DataFrame
            PSM dataframe

        ms_files : dict | list
            if dict: {raw_name: ms2 path}
            if list: [ms2 path1, ms2 path2]

        ms_file_type : str, optional
            Could be 'alpharaw_hdf', 'mgf' or 'thermo', 'sciex', 'alphapept_hdf'.
            Defaults to 'alphapept'.

        Returns
        -------
        Tuple:
            pd.DataFrame: psm dataframe with fragment index information.

            pd.DataFrame: fragment mz dataframe.

            pd.DataFrame: matched intensity dataframe.

            pd.DataFrame: matched mass error dataframe.
                np.inf if a fragment is not matched.

        """
        self.psm_df = psm_df

        (
            self.fragment_mz_df,
            self.matched_intensity_df,
            self.matched_mz_err_df,
        ) = self._prepare_matching_dfs()

        if isinstance(ms_files, dict):
            self._ms_file_dict = ms_files
        else:
            self._ms_file_dict = parse_ms_files_to_dict(ms_files)

        self._ms_file_type = ms_file_type

        self.ms_loader_thread_num = process_num
        # if process_num <= 1 or len(self._ms_file_dict) <= 1:
        psm_df_list = []
        for raw_name, df_group in tqdm.tqdm(self.psm_df.groupby("raw_name")):
            _df = self._match_ms2_one_raw_numba(raw_name, df_group)
            if _df is not None:
                psm_df_list.append(_df)
        # else:
        #     with mp.get_context("spawn").Pool(processes=process_num) as p:
        #         df_groupby = self.psm_df.groupby('raw_name')
        #         def gen_group_df(df_groupby):
        #             for raw_name, df_group in df_groupby:
        #                 yield (raw_name, df_group)
        #         process_bar(
        #             p.imap_unordered(
        #                 self._match_ms2_one_raw_numba,
        #                 gen_group_df(df_groupby)
        #             ), df_groupby.ngroups
        #         )

        self.psm_df = pd.concat(psm_df_list, ignore_index=True)
        return (
            self.psm_df,
            self.fragment_mz_df,
            self.matched_intensity_df,
            self.matched_mz_err_df,
        )


class PepSpecMatch_DIA(PepSpecMatch):
    """
    Peak annotation for DIA data.
    """

    max_spec_per_query: int = 3
    min_frag_mz: float = 200.0

    def _add_missing_columns_to_psm_df(self, psm_df: pd.DataFrame, raw_data=None):
        # DIA results do not have spec_idx/scan_num in psm_df, nothing to merge
        return psm_df

    def _prepare_matching_dfs(self):
        fragment_mz_df = self.get_fragment_mz_df()
        fragment_mz_df = pd.concat(
            [fragment_mz_df] * self.max_spec_per_query, ignore_index=True
        )
        if self.use_ppm:
            self.all_frag_mz_tols = fragment_mz_df.values * self.tolerance * 1e-6
        else:
            self.all_frag_mz_tols = np.full_like(fragment_mz_df.values, self.tolerance)

        psm_df_list = []
        len_frags = len(fragment_mz_df) // self.max_spec_per_query
        for i in range(self.max_spec_per_query):
            psm_df = self.psm_df.copy()
            psm_df["frag_start_idx"] = psm_df.frag_start_idx + i * len_frags
            psm_df["frag_stop_idx"] = psm_df.frag_stop_idx + i * len_frags
            psm_df_list.append(psm_df)
        self.psm_df = pd.concat(psm_df_list, ignore_index=True)

        matched_intensity_df = pd.DataFrame(
            np.zeros_like(fragment_mz_df.values, dtype=PEAK_INTENSITY_DTYPE),
            columns=fragment_mz_df.columns,
        )

        matched_mz_err_df = pd.DataFrame(
            np.zeros_like(fragment_mz_df.values, dtype=PEAK_MZ_DTYPE),
            columns=fragment_mz_df.columns,
        )

        return (fragment_mz_df, matched_intensity_df, matched_mz_err_df)

    def _match_ms2_one_raw_numba(
        self, raw_name: str, psm_df_one_raw: pd.DataFrame
    ) -> pd.DataFrame:
        """
        Internal method to extract peak information with numba as backend.

        Parameters
        ----------
        raw_name : str
            The raw name of the raw file. `psm_df_one_raw` dataframe should also
            contain the same raw name in `raw_name` column.
        psm_df_one_raw : pd.DataFrame
            The dataframe for PSMs.

        Returns
        -------
        pd.DataFrame
            `psm_df_one_raw`
        """
        psm_df_one_raw = psm_df_one_raw.reset_index(drop=True)

        if raw_name in self._ms_file_dict:
            raw_data = load_ms_data(
                self._ms_file_dict[raw_name],
                self._ms_file_type,
                process_count=self.ms_loader_thread_num,
            )
            if raw_data is None:
                return

            psm_origin_len = len(psm_df_one_raw) // self.max_spec_per_query

            grouper = NormalDIAGrouper(raw_data)

            psm_groups = grouper.assign_dia_groups(
                psm_df_one_raw.precursor_mz.values[:psm_origin_len]
            )

            all_spec_idxes = np.full(len(psm_df_one_raw), -1, dtype=np.int32)

            for dia_group, group_df in grouper.dia_group_dfs:
                psm_idxes = psm_groups[dia_group]
                if len(psm_idxes) == 0:
                    continue
                psm_idxes = np.array(psm_idxes, dtype=np.int32)
                spec_idxes = find_dia_spec_idxes_same_window(
                    group_df.rt.values,
                    psm_df_one_raw.rt.values[psm_idxes],
                    max_spec_per_query=self.max_spec_per_query,
                )
                for i in range(spec_idxes.shape[-1]):
                    all_spec_idxes[psm_idxes + psm_origin_len * i] = spec_idxes[:, i]

                match_one_raw_with_numba(
                    all_spec_idxes,
                    psm_df_one_raw.frag_start_idx.values,
                    psm_df_one_raw.frag_stop_idx.values,
                    self.fragment_mz_df.values,
                    self.all_frag_mz_tols,
                    raw_data.peak_df.mz.values,
                    raw_data.peak_df.intensity.values,
                    group_df.peak_start_idx.values,
                    group_df.peak_stop_idx.values,
                    self.matched_intensity_df.values,
                    self.matched_mz_err_df.values,
                    self.match_closest,
                )
        else:
            print(f"`{raw_name}` is not found in ms_file_dict.")
            return
        return psm_df_one_raw

    def match_ms2_multi_raw(
        self,
        psm_df: pd.DataFrame,
        ms_files: Tuple[dict, list],
        ms_file_type: str = "alpharaw_hdf",
        process_num: int = 8,
    ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """
        Match peaks for the given `psm_df` against the corresponding MS spectrum files.

        Parameters
        ----------
        psm_df : pd.DataFrame
            Peptide-spectrum matches in alphabase dataframe format.
        ms_files : Tuple[dict, list]
            The absolute or relative paths of MS files.
            if the type is `dict`, the format will be
            `{'raw_name1': 'raw_name1.raw', ...}` if `ms_file_type` is `thermo_raw`.
        ms_file_type : str, optional
            MS file type that is already registered in
            :obj:`alpharaw.ms_data_base.ms_reader_provider`.
            By default "alpharaw_hdf".
        process_num : int, optional
            Match peaks by using multiprocessing, by default 8

        Returns
        -------
        Tuple
            pd.DataFrame: the `psm_df`.

            pd.DataFrame: fragment m/z dataframe in alphabase format.

            pd.DataFrame: the matched fragment intensity dataframe in alphabase format.

            pd.DataFrame: the matched mass error in the same dataframe format.
        """
        if isinstance(ms_files, list):
            ms_files = parse_ms_files_to_dict(ms_files)
        psm_df = psm_df[psm_df.raw_name.isin(ms_files)].reset_index(drop=True)
        super().match_ms2_multi_raw(psm_df, ms_files, ms_file_type, process_num)

        return (
            self.psm_df,
            self.fragment_mz_df,
            self.matched_intensity_df,
            self.matched_mz_err_df,
        )


@numba.jit(nogil=True)
def match_one_raw_with_numba(
    spec_idxes: np.ndarray,
    frag_start_idxes: np.ndarray,
    frag_stop_idxes: np.ndarray,
    all_frag_mzs: np.ndarray,
    all_frag_mz_tols: np.ndarray,
    all_spec_mzs: np.ndarray,
    all_spec_intensities: np.ndarray,
    peak_start_idxes: np.ndarray,
    peak_stop_idxes: np.ndarray,
    matched_intensities: np.ndarray,
    matched_mz_errs: np.ndarray,
    match_closest: bool = True,
) -> None:
    """
    Internel function to match fragment mz values to spectrum mz values.
    Matched_mz_errs[i] = np.inf if no peaks are matched.

    Results will saved in place of matched_intensities
    and matched_mz_errs.
    """
    for spec_idx, frag_start, frag_stop in zip(
        spec_idxes, frag_start_idxes, frag_stop_idxes
    ):
        if spec_idx == -1:
            continue
        peak_start = peak_start_idxes[spec_idx]
        peak_stop = peak_stop_idxes[spec_idx]
        if peak_stop == peak_start:
            continue
        spec_mzs = all_spec_mzs[peak_start:peak_stop]
        spec_intens = all_spec_intensities[peak_start:peak_stop]

        frag_mzs = all_frag_mzs[frag_start:frag_stop, :].copy()
        frag_mz_tols = all_frag_mz_tols[frag_start:frag_stop, :].copy()

        if match_closest:
            matched_idxes = match_closest_peaks(
                spec_mzs, spec_intens, frag_mzs, frag_mz_tols
            ).reshape(-1)
        else:
            matched_idxes = match_highest_peaks(
                spec_mzs, spec_intens, frag_mzs, frag_mz_tols
            ).reshape(-1)

        matched_intens = spec_intens[matched_idxes]
        matched_intens[matched_idxes == -1] = 0

        matched_mass_errs = np.abs(
            spec_mzs[matched_idxes.reshape(-1)] - frag_mzs.reshape(-1)
        )
        matched_mass_errs[matched_idxes == -1] = np.inf

        matched_intensities[frag_start:frag_stop, :] = matched_intens.reshape(
            frag_mzs.shape
        )

        matched_mz_errs[frag_start:frag_stop, :] = matched_mass_errs.reshape(
            frag_mzs.shape
        )


def load_ms_data(
    ms_file: Union[str, MSData_Base],
    ms_file_type: str = "alpharaw_hdf",
    process_count: int = 8,
) -> MSData_Base:
    """
    Load MS file and get `MSData_Base` object.

    Parameters
    ----------
    ms_file : str | MSData_Base
        ms2 file path

    ms_file_type : str, optional
        ms2 file type, can be
        ["alpharaw_hdf", "thermo", "sciex", "alphapept_hdf", "mgf"].
        Default to 'alpharaw_hdf'.

    Returns
    -------
    MSData_Base:
        Instance of sub-class of `MSData_Base`.
    """
    if isinstance(ms_file, MSData_Base):
        return ms_file
    else:
        raw_data = ms_reader_provider.get_reader(
            ms_file_type, process_count=process_count
        )
        if raw_data is None:
            print(
                f"Raw file type '{ms_file_type}' is not registered in 'ms_reader_provider'"
            )
            return
        raw_data.import_raw(ms_file)
        return raw_data


@numba.njit
def get_best_matched_intens(
    matched_intensity_values: np.ndarray,
    frag_start_idxes: np.ndarray,
    frag_stop_idxes: np.ndarray,
):
    """
    TODO Deprecated
    """
    ret_intens = np.zeros(
        shape=matched_intensity_values.shape[1:], dtype=matched_intensity_values.dtype
    )
    for i in range(len(frag_start_idxes)):
        start = frag_start_idxes[i]
        stop = frag_stop_idxes[i]
        i = np.argmax(
            np.matmul(
                matched_intensity_values[:, start:stop, :],
                matched_intensity_values[:, start:stop, :].T,
            ).sum()
        )
        ret_intens[start:stop, :] = matched_intensity_values[i, start:stop, :]
    return ret_intens


@numba.njit
def get_ion_count_scores(
    frag_mz_values: np.ndarray,
    matched_intens: np.ndarray,
    frag_start_idxes: np.ndarray,
    frag_stop_idxes: np.ndarray,
    min_mz: float = 200,
):
    """
    TODO Deprecated
    """
    scores = []
    for i in range(len(frag_start_idxes)):
        scores.append(
            np.count_nonzero(
                matched_intens[frag_start_idxes[i] : frag_stop_idxes[i], :]
                .copy()
                .reshape(-1)[
                    frag_mz_values[frag_start_idxes[i] : frag_stop_idxes[i], :]
                    .copy()
                    .reshape(-1)
                    >= min_mz
                ]
            )
        )
    return np.array(scores, np.int32)



import os

_SPECIAL_MS_EXTS: list = [
    ".ms_data.hdf",  # alphapept
    ".raw.hdf",  # alpharaw
    ".raw.hdf5",  # alpharaw
    ".tims.hdf",  # alphatims (from alpharaw)
    ".tims.hdf5",  # alphatims (from alpharaw)
    ".d.hdf",  # alphatims (from alpharaw)
    ".d.hdf5",  # alphatims (from alpharaw)
    ".atms.hdf",  # alphatims (from alpharaw)
    ".atms.hdf5",  # alphatims (from alpharaw)
    "_hcdft.mgf",  # p
]


def get_raw_name(ms_file: str) -> str:
    """
    Get `raw_name` (base name of RAW data file) from the MS file path
    by removing the extensions defined in :data:`_SPECIAL_MS_EXTS`.

    Parameters
    ----------
    ms_file : str
        The absolute or relative path of the RAW file.

    Returns
    -------
    str
        The `raw_name` without extension.

    Examples
    --------
    >>> get_raw_name("/MS/files/your_raw_name.raw")
    'your_raw_name'

    """
    raw_name = os.path.basename(ms_file)
    lower_name = raw_name.lower()
    for _ext in _SPECIAL_MS_EXTS:
        if lower_name.endswith(_ext.lower()):
            raw_name = raw_name[: -len(_ext)]
            break
    if len(raw_name) == len(lower_name):
        raw_name = os.path.splitext(raw_name)[0]
    return raw_name


def parse_ms_files_to_dict(
    ms_file_list: list,
) -> dict:
    """
    Parse spectrum file paths into a dict:
        "/Users/xxx/raw_name1.raw" -> {"raw_name1":"/Users/xxx/raw_name1.raw"}

    Parameters
    ----------
    spectrum_file_list : list
        File path list

    special_ms_exts : list
        Special MS file extensions that contain more than one dots.

    Returns
    -------
    dict
        {"raw_name1" : "/Users/xxx/raw_name1.raw", ...}
    """

    ms_file_dict = {}
    for ms_file in ms_file_list:
        raw_name = get_raw_name(ms_file, _SPECIAL_MS_EXTS)
        ms_file_dict[raw_name] = ms_file
    return ms_file_dict



import alphatims
import numpy as np
import pandas as pd
from alphatims.bruker import TimsTOF

from ..ms_data_base import MSData_Base


class AlphaTimsReader(MSData_Base):
    """
    TimsTOF data are too large, do not use this class
    """

    def import_raw(self, burker_d_folder: str):
        tims = TimsTOF(
            burker_d_folder,
        )

        self.raw_file_path = burker_d_folder
        self.file_type = "bruker"

        self.spectrum_df["precursor_mz"] = tims.fragment_frames.IsolationMz.values
        isolations = tims.fragment_frames.IsolationWidth.values / 2
        self.spectrum_df["isolation_lower_mz"] = (
            self.spectrum_df.precursor_mz - isolations
        )
        self.spectrum_df["isolation_upper_mz"] = (
            self.spectrum_df.precursor_mz + isolations
        )
        self.spectrum_df["peak_start_idx"] = tims._push_indptr[:-1]
        self.spectrum_df["peak_stop_idx"] = tims._push_indptr[1:]
        self.spectrum_df["rt"] = tims.rt_values / 60
        self.spectrum_df["mobility"] = tims.mobility_values
        self.spectrum_df["ms_level"] = (tims.precursor_indices > 0).astype(np.int8) + 1
        self.spectrum_df["spec_idx"] = self.spectrum_df.index.values
        self.spectrum_df["tims_frame"] = (
            self.spectrum_df.spec_idx // tims.scan_max_index
        )
        self.spectrum_df["tims_scan"] = self.spectrum_df.spec_idx % tims.scan_max_index

        self.peak_df["mz"] = tims.mz_values[tims.tof_indices]
        self.peak_df["intensity"] = tims.intensity_values.astype(np.float32)


class AlphaTimsWrapper(TimsTOF):
    """Create a AlphaTims object containing
    all data in-memory (or memory mapping).

    Attribute
    ---------
    slice_as_dataframe
        Attribute from AlphaTims.
        If True, AlphaTims slicing returns a pd.DataFrame by default.
        If False, AlphaTims slicing provides a np.int64[:] with raw indices.
        The value can be modified on-the-fly.
    """

    def __init__(self, msdata: MSData_Base, dda: bool, slice_as_dataframe: bool = True):
        """
        Parameters
        ----------
        msdata : MSData_Base
            The AlphaRaw data object.

        dda : bool
            If DDA, precursor indices will be equal to scan numbers.
            If not DDA (i.e. DIA), precursor indices will be equal to the
            scan number within a DIA cycle.

        slice_as_dataframe : bool
            If True, slicing returns a pd.DataFrame by default.
            If False, slicing provides a np.int64[:] with raw indices.
            Default is True.
        """
        self._use_calibrated_mz_values_as_default = False
        self._import_alpharaw_object(msdata, dda)
        self.thermo_raw_file_name = msdata.raw_file_path
        self.bruker_d_folder_name = self.thermo_raw_file_name
        self.slice_as_dataframe = (
            slice_as_dataframe  # This value can be modified after creation.
        )
        # Precompile
        self[0, "raw"]

    def _import_alpharaw_object(
        self,
        msdata: MSData_Base,
        dda: bool,
    ):
        self._version = alphatims.__version__
        mz_values = msdata.peak_df.mz.values
        self._intensity_values = msdata.peak_df.intensity.values

        if (msdata.spectrum_df.peak_start_idx < 0).any():
            spectrum_df = msdata.spectrum_df.query("peak_start_idx!=-1")
        else:
            spectrum_df = msdata.spectrum_df

        self._push_indptr = np.zeros(len(spectrum_df) + 1, dtype=np.int64)
        self._push_indptr[1:] = spectrum_df.peak_stop_idx.values
        self._rt_values = spectrum_df.rt.values * 60
        self._quad_mz_values = spectrum_df[
            ["isolation_lower_mz", "isolation_upper_mz"]
        ].values
        if dda:
            self._precursor_indices = np.zeros_like(self._rt_values, dtype=np.int64)
            ms2s = spectrum_df.ms_level.values == 2
            self._precursor_indices[ms2s] = np.cumsum(ms2s, dtype=np.int64)[ms2s]
        else:
            precursor_indices = []
            prev_mz = -1
            prev_idx = 0
            for mz, ms_level in spectrum_df[["precursor_mz", "ms_level"]].values:
                if ms_level == 1:
                    precursor_indices.append(0)
                elif prev_mz >= mz:  # TODO if DIA mz windows are not in order
                    prev_mz = mz
                    prev_idx = 1
                    precursor_indices.append(prev_idx)
                else:
                    prev_idx += 1
                    prev_mz = mz
                    precursor_indices.append(prev_idx)
            self._precursor_indices = np.array(precursor_indices, dtype=np.int64)

        scan_count = len(self._precursor_indices)
        self._frame_max_index = scan_count
        self._scan_max_index = 1
        self._mobility_max_value = 0.0
        self._mobility_min_value = 0.0
        self._mobility_values = np.array([0.0])
        self._quad_indptr = self._push_indptr
        self._raw_quad_indptr = np.arange(scan_count + 1)
        self._intensity_min_value = float(np.min(self._intensity_values))
        self._intensity_max_value = float(np.max(self._intensity_values))
        self._intensity_corrections = np.ones(self._frame_max_index)
        _q_mzs = self._quad_mz_values[self._quad_mz_values != -1]
        if len(_q_mzs) > 0:
            self._quad_min_mz_value = float(np.min(_q_mzs))
            self._quad_max_mz_value = float(np.max(_q_mzs))
        else:
            self._quad_min_mz_value = 0
            self._quad_max_mz_value = 0
        self._precursor_max_index = int(np.max(self._precursor_indices)) + 1
        self._acquisition_mode = (
            msdata.file_type + " " + ("DDA" if dda else "DIA")
        )  # TODO
        self._mz_min_value = int(np.min(mz_values))
        self._mz_max_value = int(np.max(mz_values)) + 1
        self._decimals = 4
        self._mz_values = (
            np.arange(
                10**self._decimals * self._mz_min_value,
                10**self._decimals * (self._mz_max_value + 1),
            )
            / 10**self._decimals
        )
        self._tof_indices = (mz_values * 10**self._decimals).astype(
            np.int32
        ) - 10**self._decimals * self._mz_min_value
        self._tof_max_index = len(self._mz_values)
        self._meta_data = {"SampleName": msdata.raw_file_path}
        msmstype = np.array([0 if s == -1 else 1 for s, e in self._quad_mz_values])
        summed_intensities_ = np.cumsum(self._intensity_values)
        summed_intensities = -summed_intensities_[self._push_indptr[:-1]]
        summed_intensities[:-1] += summed_intensities_[self._push_indptr[1:-1]]
        summed_intensities[-1] += summed_intensities_[-1]
        max_intensities = [
            np.max(
                self._intensity_values[self._push_indptr[i] : self._push_indptr[i + 1]]
            )
            if self._push_indptr[i + 1] != -1
            and self._push_indptr[i] != -1
            and self._push_indptr[i] != self._push_indptr[i + 1]
            else 0
            for i in range(len(self._rt_values))
        ]
        self._frames = pd.DataFrame(
            {
                "MsMsType": msmstype,
                "Time": self._rt_values,
                "SummedIntensities": summed_intensities,
                "MaxIntensity": max_intensities,
                "Id": np.arange(len(self._rt_values)),
            }
        )
        frame_numbers = np.arange(len(self._rt_values), dtype=np.int32)
        isolation_widths = self._quad_mz_values[:, 1] + self._quad_mz_values[:, 0]
        isolation_centers = self._quad_mz_values[:, 1] - self._quad_mz_values[:, 0]
        self._fragment_frames = pd.DataFrame(
            {
                "Frame": frame_numbers[msmstype == 1],
                "ScanNumBegin": 0,
                "ScanNumEnd": 0,
                "IsolationWidth": isolation_widths[msmstype == 1],
                "IsolationMz": isolation_centers[msmstype == 1],
                "Precursor": self._precursor_indices[msmstype == 1],
            }
        )
        self._zeroth_frame = False
        offset = int(self.zeroth_frame)
        cycle_index = (
            np.searchsorted(
                self.raw_quad_indptr,
                (self.scan_max_index) * (self.precursor_max_index + offset),
                "r",
            )
            + 1
        )
        repeats = np.diff(self.raw_quad_indptr[:cycle_index])
        if self.zeroth_frame:
            repeats[0] -= self.scan_max_index
        cycle_length = self.scan_max_index * self.precursor_max_index
        repeat_length = np.sum(repeats)
        if repeat_length != cycle_length:
            repeats[-1] -= repeat_length - cycle_length
        self._dia_mz_cycle = np.empty((cycle_length, 2))
        self._dia_mz_cycle[:, 0] = np.repeat(
            self.quad_mz_values[: cycle_index - 1, 0], repeats
        )
        self._dia_mz_cycle[:, 1] = np.repeat(
            self.quad_mz_values[: cycle_index - 1, 1], repeats
        )
        self._dia_precursor_cycle = np.repeat(
            self.precursor_indices[: cycle_index - 1], repeats
        )
